
# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation
"""Provides functions to build MNE objects (raw, epoch, evoked) from PyRates simulation results (csv) or from circuit
objects.
"""

# external packages
import numpy as np
from typing import Union, Optional, List
from pandas import DataFrame, MultiIndex

# pyrates internal imports

# meta infos
__author__ = 'Richard Gast'
__status__ = 'Development'


################################################################################
# wrapper to build a pyrates-compatible output structure from annarchy monitor #
################################################################################


def pyrates_from_annarchy(monitors: list, vars: List[str], pop_average: bool = False,
                          monitor_names: Optional[List[str]] = None, **kwargs) -> DataFrame:
    """

    Parameters
    ----------
    monitors
    vars
    pop_average
    monitor_names

    Returns
    -------
    DataFrame

    """

    if not monitor_names:
        monitor_names = [str(i) for i in range(len(monitors))]
    results = {}
    for v in vars:
        for m, m_name in zip(monitors, monitor_names):
            if len(m_name) == 0:
                m_name = v
            if v == 'spike':
                val = m.get(v)
                val = m.smoothed_rate(val, **kwargs).T
            else:
                val = np.asarray(m.get(v))
            if len(val.shape) > 1:
                if pop_average:
                    results.update({f'{m_name}_avg': np.mean(val, axis=1)})
                else:
                    results.update({f'{m_name}_n{i}': val[:, i] for i in range(val.shape[1])})
            else:
                results.update({f'{m_name}': val})
    return DataFrame.from_dict(results)


#################################################
# grid search function for annarchy simulations #
#################################################


def grid_search_annarchy(param_grid: dict, param_map: dict, dt: float, simulation_time: float,
                         inputs: dict, outputs: dict, sampling_step_size: Optional[float] = None,
                         permute_grid: bool = False, circuit=None, **kwargs) -> DataFrame:
    """Function that runs multiple parametrizations of the same circuit in parallel and returns a combined output.

    Parameters
    ----------
    param_grid
        Key-value pairs for each circuit parameter that should be altered over different circuit parametrizations.
    param_map
        Key-value pairs that map the keys of param_grid to concrete circuit variables.
    dt
        Simulation step-size in s.
    simulation_time
        Simulation time in s.
    inputs
        Inputs as provided to the `run` method of `:class:ComputeGraph`.
    outputs
        Outputs as provided to the `run` method of `:class:ComputeGraph`.
    sampling_step_size
        Sampling step-size as provided to the `run` method of `:class:ComputeGraph`.
    permute_grid
        If true, all combinations of the provided param_grid values will be realized. If false, the param_grid values
        will be traversed pairwise.
    circuit
        Instance of ANNarchy network.
    kwargs
        Additional keyword arguments passed to the `:class:ComputeGraph` initialization.



    Returns
    -------
    DataFrame
        Simulation results stored in a multi-index data frame where each index lvl refers to one of the parameters of
        param_grid.

    """

    from ANNarchy import Population, Projection, Network, TimedArray, Monitor, ANNarchyException

    # linearize parameter grid if necessary
    if type(param_grid) is dict:
        param_grid = linearize_grid(param_grid, permute_grid)

    # create annarchy net if necessary
    if circuit is None:
        circuit = Network(everything=True)

    # assign parameter updates to each circuit and combine them to unconnected network
    circuit_names = []
    param_info = []
    param_split = "__"
    val_split = "--"
    comb = "_"
    populations, projections = {}, {}
    for n in range(param_grid.shape[0]):

        # copy and re-parametrize populations
        try:
            for p in circuit.get_populations():
                name = f'net{n}/{p.name}'
                p_new = Population(geometry=p.geometry, neuron=p.neuron_type, name=name,
                                   stop_condition=p.stop_condition, storage_order=p._storage_order,
                                   copied=False)
                p_new = adapt_pop(p_new, param_grid.iloc[n, :], param_map)
                populations[name] = p_new

                # add input to population
                for node, inp in inputs.items():
                    if node in name:
                        inp_name = f'{name}_inp'
                        inp = TimedArray(rates=inp, name=inp_name)
                        proj = Projection(pre=inp, post=p_new, target='exc')
                        proj.connect_one_to_one(1.0)
                        populations[inp_name] = inp
                        projections[inp_name] = proj
        except ANNarchyException:
            pass

        # copy and re-parametrize projections
        try:
            for c in circuit.get_projections():
                source = c.pre if type(c.pre) is str else c.pre.name
                target = c.post if type(c.post) is str else c.post.name
                source = f'net{n}/{source}'
                target = f'net{n}/{target}'
                name = f'{source}/{target}/{c.name}'
                c_new = Projection(pre=source, post=target, target=c.target, synapse=c.synapse_type, name=name,
                                   copied=False)
                c_new._store_connectivity(c._connection_method, c._connection_args, c._connection_delay, c._storage_format)
                c_new = adapt_proj(c_new, param_grid.iloc[n, :], param_map)
                projections[name] = c_new
        except ANNarchyException:
            pass

        # collect parameter and circuit name infos
        circuit_names.append(f'net{n}')
        param_names = list(param_grid.columns.values)
        param_info_tmp = [f"{param_names[i]}{val_split}{val}" for i, val in enumerate(param_grid.iloc[n, :])]
        param_info.append(param_split.join(param_info_tmp))

    net = Network()
    for p in populations.values():
        net.add(p)
    for c in projections.values():
        net.add(c)

    # adjust output of simulation to combined network
    nodes = [p.name for p in circuit.get_populations()]
    out_names, var_names, out_lens, monitors, monitor_names = [], [], [], [], []
    for out_key, out in outputs.copy().items():
        out_names_tmp, out_lens_tmp = [], []
        if out[0] in nodes:
            for i, name in enumerate(param_info):
                out_tmp = list(out)
                out_tmp[0] = f'{circuit_names[i]}/{out_tmp[0]}'
                p = net.get_population(out_tmp[0])
                monitors.append(Monitor(p, variables=out_tmp[-1], period=sampling_step_size, start=True,
                                        net_id=net.id))
                monitor_names.append(f'{name}{param_split}out_var{val_split}{out_key}{comb}{out[0]}')
                var_names.append(out_tmp[-1])
                out_names_tmp.append(f'{out_key}{comb}{out[0]}')
                out_lens_tmp.append(p.geometry[0])
        elif out[0] == 'all':
            for node in nodes:
                for i, name in enumerate(param_info):
                    out_tmp = list(out)
                    out_tmp[0] = f'{circuit_names[i]}/{node}'
                    p = net.get_population(out_tmp[0])
                    monitors.append(Monitor(p, variables=out_tmp[-1], period=sampling_step_size, start=True,
                                            net_id=net.id))
                    monitor_names.append(f'{name}{param_split}out_var{val_split}{out_key}{comb}{node}')
                    var_names.append(out_tmp[-1])
                    out_names_tmp.append(f'{out_key}{comb}{node}')
                    out_lens_tmp.append(p.geometry[0])
        else:
            node_found = False
            for node in nodes:
                if out[0] in node:
                    node_found = True
                    for i, name in enumerate(param_info):
                        out_tmp = list(out)
                        out_tmp[0] = f'{circuit_names[i]}/{node}'
                        p = net.get_population(out_tmp[0])
                        monitors.append(Monitor(p, variables=out_tmp[-1], period=sampling_step_size, start=True,
                                                net_id=net.id))
                        monitor_names.append(f'{name}{param_split}out_var{val_split}{out_key}{comb}{node}')
                        var_names.append(out_tmp[-1])
                        out_names_tmp.append(f'{out_key}{comb}{node}')
                        out_lens_tmp.append(p.geometry[0])
            if not node_found:
                raise ValueError(f'Invalid output identifier in output: {out_key}. '
                                 f'Node {out[0]} is not part of this network')
        out_names += list(set(out_names_tmp))
        out_lens += list(set(out_lens_tmp))
    #net.add(monitors)

    # simulate the circuits behavior
    net.compile()
    net.simulate(duration=simulation_time)

    # transform output into pyrates-compatible data format
    results = pyrates_from_annarchy(monitors, vars=list(set(var_names)),
                                    monitor_names=monitor_names, **kwargs)

    # transform results into long-form dataframe with changed parameters as columns
    multi_idx = [param_grid[key].values for key in param_grid.keys()]
    n_iters = len(multi_idx[0])
    outs = []
    for out_name, out_len in zip(out_names, out_lens):
        outs += [f'{out_name}_n{i}' for i in range(out_len)] * n_iters
    multi_idx_final = []
    for idx in multi_idx:
        for val in idx:
            for out_len in out_lens:
                multi_idx_final += [val]*len(out_names)*out_len
    index = MultiIndex.from_arrays([multi_idx_final, outs], names=list(param_grid.keys()) + ["out_var"])
    index = MultiIndex.from_tuples(list(set(index)), names=list(param_grid.keys()) + ["out_var"])
    results_final = DataFrame(columns=index, data=np.zeros_like(results.values), index=results.index)
    for col in results.keys():
        params = col.split(param_split)
        indices = [None] * len(results_final.columns.names)
        for param in params:
            var, val = param.split(val_split)[:2]
            idx = list(results_final.columns.names).index(var)
            try:
                indices[idx] = float(val)
            except ValueError:
                indices[idx] = val
        results_final.loc[:, tuple(indices)] = results[col].values

    return results_final


##########################################
# utility functions for annarchy objects #
##########################################


def adapt_pop(pop, params: dict, param_map: dict):
    """Changes the parametrization of a circuit.

    Parameters
    ----------
    pop
        ANNarchy population instance.
    params
        Key-value pairs of the parameters that should be changed.
    param_map
        Map between the keys in params and the circuit variables.

    Returns
    -------
    Population
        Updated population instance.

    """

    for key in params.keys():
        val = params[key]
        for op, var in param_map[key]['var']:
            nodes = param_map[key]['nodes'] if 'nodes' in param_map[key] else []
            for node in nodes:
                if node in pop.name:
                    try:
                        pop.set({var: float(val)})
                    except TypeError:
                        pop.set({var: val})
                    except (KeyError, ValueError):
                        pass

    return pop


def adapt_proj(proj, params: dict, param_map: dict):
    """Changes the parametrization of a circuit.

    Parameters
    ----------
    proj
        ANNarchy projection instance.
    params
        Key-value pairs of the parameters that should be changed.
    param_map
        Map between the keys in params and the circuit variables.

    Returns
    -------
    Projection
        Updated projection instance.

    """

    for key in params.keys():
        val = params[key]
        for op, var in param_map[key]['var']:
            edges = param_map[key]['edges'] if 'edges' in param_map[key] else []
            for source, target, edge in edges:
                if source in proj.name and target in proj.name and edge in proj.name:
                    try:
                        proj.set({var: float(val)})
                    except TypeError:
                        proj.set({var: val})
                    except [ValueError, KeyError]:
                        pass

    return proj


def linearize_grid(grid: dict, permute: bool = False) -> DataFrame:
    """Turns the grid into a grid that can be traversed linearly, i.e. pairwise.

    Parameters
    ----------
    grid
        Parameter grid.
    permute
        If true, all combinations of the parameter values in grid will be created.

    Returns
    -------
    DataFrame
        Resulting linear grid in form of a data frame.

    """

    arg_lengths = [len(arg) for arg in grid.values()]

    if len(list(set(arg_lengths))) == 1 and not permute:
        return DataFrame(grid)
    else:
        vals, keys = [], []
        for key, val in grid.items():
            vals.append(val)
            keys.append(key)
        new_grid = np.stack(np.meshgrid(*tuple(vals)), -1).reshape(-1, len(grid))
        return DataFrame(new_grid, columns=keys)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from collections import namedtuple as _namedtuple
from copy import deepcopy
from typing import List, Iterator

from pyrates.ir.abc import AbstractBaseIR

__author__ = "Daniel Rose"
__status__ = "Development"

Variable = _namedtuple("Variable", ["vtype", "dtype", "shape"])


class ProtectedVariableDict:
    """Hashable dictionary-like object as container for variable definitions. It does not support item assignment after
    creation, but is strictly speaking not immutable. There may also be faster implementations, but this works for
    now."""

    __slots__ = ["_hash", "_d", "_parsed"]

    def __init__(self, variables: List[tuple]):
        variables = tuple(variables)
        self._hash = hash(variables)
        self._d = {vname: dict(vtype=vtype, dtype=dtype, shape=shape)
                   for vname, vtype, dtype, shape in variables}

    def add_parsed_variable(self, key, props):
        """Add parsed representation to a variable from compilation."""
        self._parsed[key] = props

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        return self._hash

    def items(self):
        return self._d.items()

    def keys(self):
        return self._d.keys()

    def values(self):
        return self._d.values()

    def to_dict(self):
        return deepcopy(self._d)

# class ProtectedVariableDict(dict):
#     """This is an unsafe hack to provide an immutable and hashable dict. Unsafe means, that checks against isinstance
#     or
#     issubtype might lead to wrong conclusions about the mutability. There may also be faster implementations, but this
#     works for now."""
#
#     def __init__(self, variables: List[tuple]):
#         variables = tuple(variables)
#         self._h = hash(variables)
#
#         named_variables = {vname: Variable(vtype, dtype, shape)
#                            for vname, vtype, dtype, shape in variables}
#
#         super().__init__(**named_variables)
#
#     def __setitem__(self, key, value):
#         raise TypeError("'VariableDict' object does not support item assignment")
#
#     # def __setattr__(self, key, value):
#     #     raise AttributeError(f"'VariableDict' object has not attribute '{key}'")
#
#     def __hash__(self):
#         return self._h


class OperatorIR(AbstractBaseIR):
    """This implementation of the Operator IR is aimed to be hashable and immutable. Following Python standards, we
    assume that users are consenting adults. Objects are thus not actually immutable, just slightly protected.
    This might change in the future."""
    __slots__ = ["_equations", "_variables", "_inputs", "_output"]

    def __init__(self, equations: List[str], variables: List[tuple], inputs: List[str], output: str,
                 template: str = None):

        super().__init__(template)
        # define hash

        self._equations = tuple(equations)

        self._variables = ProtectedVariableDict(variables)
        self._inputs = tuple(inputs)
        self._output = output
        self._h = hash((self._equations, self._variables, self._inputs, self._output))

    @property
    def variables(self):
        return self._variables

    @property
    def equations(self):
        return self._equations

    @property
    def inputs(self):
        return self._inputs

    @property
    def output(self):
        return self._output

    # @staticmethod
    # def _reduce_ode_order(equations: str, variables):
    #     """Checks if a 2nd-order ODE is present and reduces it to two coupled first-order ODEs.
    #     Currently limited to special case of the form '(d/dt + a)^2 * x = b'.
    #
    #     Parameters
    #     ----------
    #     equations
    #         string of form 'a = b'
    #     """
    #
    #     # matches pattern of form `(d/dt + a)^2 * y` and extracts `a` and `y`
    #     match = re.match(r"\(\s*d\s*/\s*dt\s*[+-]\s*([\d]*[.]?[\d]*/?[a-zA-Z]\w*)\s*\)\s*\^2\s*\*\s*([a-zA-Z]\w*)",
    #                      equations)
    #
    #     if match:
    #         # assume the entire lhs was matched, fails if there is something remaining on the lhs
    #         lhs, rhs = equations.split("=")
    #         a, var = match.groups()  # returns coefficient `a` and variable `y`
    #         eq1 = f"d/dt * {var} = {var}_t"
    #         eq2 = f"d/dt * {var}_t = {rhs} - ({a})^2 * {var} - 2. * {a} * {var}_t"
    #
    #         variables[f"{var}_t"] = {"dtype": "float32",
    #                                  "description": "integration variable",
    #                                  "vtype": "state_var",
    #                                  "value": variables[var]['value'] if variables[var]['value'] else 0.}
    #
    #         return eq1, eq2, variables
    #     else:
    #         return equations, variables

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """
        Checks if a variable named by key exists in an equations.
        Parameters
        ----------
        key
        key_iter

        Returns
        -------
        key
        """

        for equation in self.equations:
            if key in equation:
                return key
        else:
            raise KeyError(f"Variable `{key}` not found in equations {self.equations}")

    def __str__(self):

        return f"<{self.__class__.__name__}({self.equations}), hash = {hash(self)} >"




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" This module contains classes relevant for the intermediate representation (IR) in PyRates. All variations of PyRates
frontends should produce valid instances of IR classes defined in this module and all backends should accept only these
instances as input. Thus the intermediate representation is assumed to be consistent across different frontends and
backends.
"""

__author__ = "Daniel Rose"
__status__ = "Development"

from .circuit import CircuitIR

_unique_objects = dict()


# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from copy import copy, deepcopy
from typing import Iterator

import numpy as np

from pyrates.ir.abc import AbstractBaseIR
from pyrates.ir.operator_graph import OperatorGraph, VectorizedOperatorGraph

__author__ = "Daniel Rose"
__status__ = "Development"


class NodeIR(AbstractBaseIR):
    __slots__ = ["_op_graph", "values"]

    def __init__(self, operators: dict = None, values: dict = None, template: str = None):
        super().__init__(template)
        self._op_graph = OperatorGraph(operators)
        self.values = values

    @property
    def op_graph(self):
        return self._op_graph

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """Alias for self.op_graph.getitem_from_iterator"""

        return self.op_graph.getitem_from_iterator(key, key_iter)

    def __iter__(self):
        """Return an iterator containing all operator labels in the operator graph."""
        return iter(self.op_graph)

    @property
    def operators(self):
        return self.op_graph.operators

    def __hash__(self):
        raise NotImplementedError


class VectorizedNodeIR(AbstractBaseIR):
    """Alternate version of NodeIR that takes a full NodeIR as input and creates a vectorized form of it."""

    __slots__ = ["op_graph", "_length"]

    def __init__(self, node_ir: NodeIR):

        super().__init__(node_ir.template)
        self.op_graph = VectorizedOperatorGraph(node_ir.op_graph, node_ir.values)
        values = {}
        # reformat all values to be lists of themselves (adding an outer vector dimension)
        # if len(node_ir.op_graph) == 0:
        #     op_key, data = next(iter(self.op_graph.node(data=True)))
        #     for var in data["variables"]:
        #         values[op_key] = {var: [0.]}
        # else:

        # save current length of this node vector.
        self._length = 1

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """Alias for self.op_graph.getitem_from_iterator"""

        return self.op_graph.getitem_from_iterator(key, key_iter)

    def __iter__(self):
        """Return an iterator containing all operator labels in the operator graph."""
        return iter(self.op_graph)

    @property
    def operators(self):
        return self.op_graph.operators

    def __hash__(self):
        raise NotImplementedError

    def extend(self, node: NodeIR):
        """ Extend variables vectors by values from one additional node.

        Parameters
        ----------
        node
            A node whose values are used to extend the vector dimension of this vectorized node.

        Returns
        -------
        """

        # add values to respective lists in collapsed node
        self.op_graph.append_values(node.values)

        self._length += 1

    def __len__(self):
        """Returns size of this vector node as recorded in self._length.

        Returns
        -------
        self._length
        """
        return self._length

    def add_op(self, op_key: str, inputs: dict, output: str, equations: list, variables: dict):
        """Wrapper for internal `op_graph.add_operator` that adds any values to node-level values dictionary for quick
        access

        Parameters
        ----------
        op_key
            Name of operator to be added
        inputs
            dictionary definining input variables of the operator
        output
            string defining name of single output variable
        equations
            list of equations (strings)
        variables
            dictionary describing variables

        Returns
        -------

        """

        # add operator to op_graph
        self.op_graph.add_operator(op_key, inputs=inputs, output=output, equations=equations, variables=variables)

    def add_op_edge(self, source_op_key: str, target_op_key: str, **attr):
        """ Alias to `self.op_graph.add_edge`

        Parameters
        ----------
        source_op_key
        target_op_key
        attr

        Returns
        -------

        """

        self.op_graph.add_edge(source_op_key, target_op_key, **attr)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

# external imports
import numpy as np
import pandas as pd
from typing import Optional, Union

# system imports
import os
import h5py
import random
from itertools import cycle

from pyrates.utility.grid_search import grid_search, ClusterGridSearch, linearize_grid

# meta infos
__author__ = "Christoph Salomon, Richard Gast"
__status__ = "development"


class GeneticAlgorithmTemplate:
    def __init__(self):

        # Initialize storage variables
        self.initial_gene_pool = {}
        self.num_genes = 0
        self.sigma_adapt = 0
        self.gene_names = []
        self.pop = pd.DataFrame()
        self.pop_size = 0
        self.candidate = pd.DataFrame()
        self.winner = pd.DataFrame()
        self.current_max_fitness = 0
        self.drop_count = 0

    def run(self, initial_gene_pool: dict, target: list, max_iter: int, min_fit: Optional[float] = 0.,
            n_winners: Optional[int] = 1, n_parent_pairs: Optional[int] = 10, n_new: Optional[int] = 0,
            sigma_adapt: Optional[float] = 0., max_stagnation_steps: Optional[int] = 0,
            stagnation_decimals: Optional[int] = 8, max_stagnation_drops: Optional[Union[int, float]] = np.Inf,
            enforce_max_iter: Optional[bool] = False, new_pop_on_drop: Optional[bool] = False,
            candidate_save: Optional[str] = "", drop_save: Optional[str] = "", gene_sampling_func=np.linspace, **kwargs):
        """Run a genetic algorithm to optimize genes of a population in respect to a given target vector

        Parameters
        ----------
        initial_gene_pool
            Dictionary containing ranges for each gene to sample from
        target
            Target values that are used to determine the fitness of a population member
        max_iter
            Maximum number of iterations (generations) before the computation is terminated
        min_fit
            Minimum fitness. If set, computation will stop after the fitness of one population member exceeds this value
        n_winners
            Number of strongest members of a population, that will be members of this population's offspring
        n_parent_pairs
            Number of parent pairs of a population that will produce a child which will be member of this population's
            offspring (crossover)
        n_new
            Number of members in a population's offspring, that will be created from the initial gene pool
        sigma_adapt
            Ratio by which the mean derivation of the gene probability distributions will change during a mutation as
            suggested by (Beyer1995 - Toward a Theory of Evolution Strategies: Self-Adaptation)
        max_stagnation_steps
            Maximum number of iterations with no change in the fitness, before the strongest member of a population
            will be discarded and not be part of this population's offspring anymore
        stagnation_decimals
            Decimal precision that will be used to detect changes in the fitness of successive populations
        max_stagnation_drops
            If True, computation will stop when the maximum stagnation is reached
        enforce_max_iter
            If True, all iterations will performed, even if another convergence criterion is reached before
        new_pop_on_drop
            If True, a new population is created once the fitness stagnates. If False, only the fittest member of
            the population is replaced by a new member and the computation continues
        candidate_save
            If set, the strongest member of a population will be saved to an hdf5 file, before the population is updated
        drop_save
            If set, all members that are dropped from a population due to stagnation or other criteria will be saved to
            this folder in hdf5 format
        gene_sampling_func
        kwargs


        Returns
        -------
        pandas.DataFrame containing the overall fittest member of all computed populations

        """

        self.initial_gene_pool = initial_gene_pool
        self.num_genes = len(initial_gene_pool)
        self.sigma_adapt = sigma_adapt

        # Counts how many members have already been droped out from a population due to fitness stagnation
        self.drop_count = 0

        # Create starting population
        ############################
        self.__create_pop(sampling_func=gene_sampling_func)
        self.pop_size = self.pop.shape[0]

        if n_winners + n_parent_pairs + n_new > self.pop_size:
            print('WARNING: Sum of winners, parents and new members exceeds the population size. Returning')
            return
        # Start genetic algorithm
        #########################
        print("***STARTING GENETIC ALGORITHM***")
        iter_count = 0
        stagnation_count = 0
        while iter_count < max_iter:
            print("")
            print(f'ITERATION: {iter_count}')

            # Evaluate fitness of current population
            ########################################
            self.eval_fitness(target, **kwargs)
            new_candidate = self.pop.nlargest(1, "fitness")
            self.current_max_fitness = float(new_candidate.loc[:, "fitness"])

            # If no population member yields a proper fitness value since all computed timeseries contained at least one
            # undefined value (e.g. np.NaN)
            print(f'Currently fittest genes:')
            self.plot_genes(new_candidate)
            target_tmp = []
            for tar in target:
                if isinstance(tar, list):
                    target_tmp.append(np.round(tar[0], decimals=2))
                else:
                    target_tmp.append(np.round(tar, decimals=2))
            print(f'Target: {target_tmp}')

            # Check for fitness stagnation
            ##############################
            if max_stagnation_steps > 0:
                # Before the first iteration self.candidate is empty. Skip stagnation check in that case
                if not self.candidate.empty:
                    old_fitness = np.round(float(self.candidate['fitness']), decimals=stagnation_decimals)
                    new_fitness = np.round(self.current_max_fitness, decimals=stagnation_decimals)
                    # Check for change in fitness
                    if new_fitness == old_fitness:
                        stagnation_count += 1
                        # Check if stagnation occured
                        if stagnation_count > max_stagnation_steps:
                            print("Maximum fitness stagnation reached!")
                            # Check if maximum number of population drops is reached
                            if not (self.drop_count == max_stagnation_drops) or enforce_max_iter:
                                if drop_save:
                                    print("Saving candidate!")
                                    os.makedirs(drop_save, exist_ok=True)
                                    new_candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
                                    with h5py.File(f'{drop_save}/PopulationDrop_{self.drop_count}.h5') as file:
                                        file['target'] = target

                                self.drop_count += 1

                                if new_pop_on_drop:
                                    print("Dropping fittest from population!")
                                    self.__create_pop(sampling_func=gene_sampling_func)
                                    continue
                                else:
                                    print("Dropping candidate from population!")
                                    self.pop = self.pop.drop(new_candidate.index)
                            else:
                                print("Returning fittest member!")
                                print("")
                                return new_candidate
                    else:
                        # Reset stagnation counter
                        stagnation_count = 0

            # Update candidate and save if necessary
            ########################################
            self.candidate = new_candidate
            if candidate_save:
                self.candidate.to_hdf(candidate_save, key='data')

            # Update current winning genes
            #############################
            if self.winner.empty:
                self.winner = self.candidate
            # Cast floats since truth value of a pd.Series is ambiguous
            elif float(self.candidate['fitness']) > float(self.winner['fitness']):
                self.winner = self.candidate

            # Evaluate minimum fitness conversion criteria
            ##############################################
            if 0 < min_fit < self.current_max_fitness:
                print("Minimum fitness criterion reached!")
                if enforce_max_iter:
                    if drop_save:
                        print("Saving candidate!")
                        new_candidate.to_hdf(f'{drop_save}/PopulationDrop_{self.drop_count}.h5', key='data')
                    self.drop_count += 1
                    if new_pop_on_drop:
                        print(f'Generating new population')
                        self.__create_pop(sampling_func=gene_sampling_func)
                        continue
                    else:
                        print("Dropping candidate from population!")
                        self.pop = self.pop.drop(self.candidate.index)
                else:
                    return self.candidate

            # Create offspring from current population
            ##########################################
            if self.current_max_fitness == -0.0:
                print(f'No candidate available for the current gene set')
                print(f'Generating new population')
                self.__create_pop(sampling_func=gene_sampling_func)
                iter_count += 1
            else:
                print(f'Generating offspring')
                self.__create_offspring(n_parent_pairs=n_parent_pairs, n_new=n_new, n_winners=n_winners)
                iter_count += 1

        # End of iteration loop
        print("Maximum iterations reached")
        if float(self.winner['fitness']) < min_fit:
            print('Could not satisfy minimum fitness condition.')
        return self.winner

    def __create_offspring(self, n_winners, n_parent_pairs=0, n_new=0):
        """Create a new offspring of the current population

        Offspring contains:
        - n_winners strongest members of the current population (winners)
        - n_parent_pairs children of current parent pairings (crossover)
        - n_mut Mutations of winners and children (mutations)
        - n_new Fresh members based on the initial gene pool (new)

        The number of mutations is chosen dynamically to resize the offspring to the size of the current population
        with n_mut = population_size - n_winners - n_parent_pairs - n_new
        """
        print('Updating population')

        # Create new offspring
        ######################
        offspring = []
        new_sigs = []
        n_mutations = self.pop_size - (n_parent_pairs + n_new + n_winners)

        # 1. Add n_winners strongest members
        ####################################
        winners = self.__select_winners(n_winners=n_winners)
        for w in winners:
            offspring.append(w[0])
            new_sigs.append(w[1])

        # 2. Add children of n_parents parent pairs
        ###########################################
        parent_pairs = self.__create_parent_pairs(n_parent_pairs=n_parent_pairs)
        childs = []
        try:
            childs = self.__crossover(parent_pairs)
            for c in childs:
                offspring.append(c[0])
                new_sigs.append(c[1])
        except ValueError:
            # Each failed child will be replaced by a mutation
            n_mutations += (n_parent_pairs - len(childs))

        # 3. Add mutations
        ##################
        parent_pool = cycle(zip(offspring, new_sigs))
        for mut in range(n_mutations):
            parent = next(parent_pool)
            mutation = self.__mutate(parent)
            offspring.append(mutation[0])
            new_sigs.append(mutation[1])

        # 4. Add n_new fresh members from initial gene_pool
        ###################################################
        for n in range(n_new):
            new_member = self.__create_new_member()
            offspring.append(new_member[0])
            new_sigs.append(new_member[1])

        offspring = pd.DataFrame(offspring)

        # 5. Swap possible duplicates in the offspring with new members
        ###############################################################
        while any(offspring.duplicated()):
            dupl_idx = offspring.loc[offspring.duplicated()].index.to_numpy()
            for i in dupl_idx:
                # Replace every duplicate with a new chromosome, fitness 0.0 and respective sigmas
                new_member = self.__create_new_member()
                offspring.iloc[i] = new_member[0]
                new_sigs[i] = new_member[1]

        offspring.columns = self.pop.loc[:, self.gene_names].columns
        offspring['fitness'] = 0.0
        offspring['sigma'] = new_sigs

        self.pop = offspring

    def __create_pop(self, sampling_func=np.linspace):
        """Create new population from the initial gene pool"""
        pop_grid = {}
        # Prevent duplicates if create_pop() is called again if population had no winner
        self.gene_names = []
        for param, value in self.initial_gene_pool.items():
            self.gene_names.append(param)
            if value['N'] == 1:
                pop_grid[param] = np.array([np.mean([value['min'], value['max']])])
            else:
                pop_grid[param] = sampling_func(value['min'], value['max'], value['N'])
            # pop_grid[param] = np.linspace(value['min'], value['max'], value['N'])
        self.pop = linearize_grid(pop_grid, permute=True)
        self.pop_size = self.pop.shape[0]

        sigmas = [self.initial_gene_pool[gene]['sigma'] for gene in self.initial_gene_pool.keys()]

        self.pop['fitness'] = 0.0
        self.pop['sigma'] = [sigmas for _ in range(self.pop_size)]

    def __select_winners(self, n_winners):
        """Returns the n_winners fittest members from the current population"""
        winners = []
        for idx in self.pop.nlargest(n_winners, 'fitness').index:
            # winner_genes = self.pop.iloc[idx].drop(['fitness', 'sigma']).to_list()
            winner_genes = self.pop.loc[idx, self.gene_names].to_list()
            winner_sigma = self.pop.iloc[idx]['sigma']
            winners.append([winner_genes, winner_sigma])
        return winners

    def __mutate(self, parent):
        """Create mutation of a parent, based on a gaussian distribution for each gene"""
        mu_new = []
        sigma_new = []
        for i, (mu, sigma) in enumerate(zip(parent[0], parent[1])):
            mu_temp = mu + np.random.randn() * sigma
            while any([mu_temp < self.initial_gene_pool[self.gene_names[i]]['min'],
                       mu_temp > self.initial_gene_pool[self.gene_names[i]]['max']]):
                mu_temp = mu+np.random.randn()*sigma
            mu_new.append(mu_temp)
            # Adapt sigma (Beyer1995, p.5)
            xi = np.exp(self.sigma_adapt*np.random.randn())
            sigma_new.append(sigma*xi)
        return mu_new, sigma_new

    def __create_new_member(self):
        """Create a new population member from the initial gene pool"""
        genes = []
        sigma = []
        for i, (key, value) in enumerate(self.initial_gene_pool.items()):
            genes.append(random.uniform(value['min'], value['max']))
            sigma.append(value['sigma'])
        new_member = [genes, sigma]
        already_exists = (self.pop.loc[:, self.gene_names] == new_member[0]).all(1).any()
        if already_exists:
            new_member = self.__create_new_member()
        return new_member

    def __create_parent_pairs(self, n_parent_pairs):
        """Create n_parent_pairs parent combinations. The occurrence probability for each parent is based on that
        parent's fitness"""
        parents = []
        # Reproduction probability for each parent is based on its relative fitness
        parent_repro = self.pop['fitness'].copy()

        # Set -inf and NaN to 0 since np.choice can only handle positive floats or ints
        # Safety measure, should not occur in the first place
        parent_repro[parent_repro == -np.inf] = 0.0
        parent_repro[parent_repro == -np.Inf] = 0.0
        parent_repro[parent_repro == np.NaN] = 0.0
        parent_repro[parent_repro == np.nan] = 0.0

        # Convert fitness to list of normalized choice probabilities
        parent_repro = np.abs(parent_repro.to_numpy())
        parent_repro /= np.abs(parent_repro.sum())

        # Get a list containing the indices of all population members
        parent_indices = self.pop.index.values
        for n in range(n_parent_pairs):
            p_idx = np.random.choice(parent_indices, size=(2,), replace=False, p=parent_repro)
            parents.append((self.pop.iloc[p_idx[0], :], self.pop.iloc[p_idx[1], :]))
        return parents

    def __crossover(self, parent_pairs, n_tries=5):
        """Create a child from each parent pair. Each child gene is uniformly chosen from one of its parents

        If the child already exists in the current population, new genes are chosen, but maximal n_tries times before a
        ValueError is raised.
        """
        childs = []
        for parents in parent_pairs:
            count = 0
            while True:
                if count > n_tries:
                    raise ValueError
                child_genes = []
                child_sigma = []
                for g, gene in enumerate(self.initial_gene_pool):
                    choice = np.random.uniform()
                    if choice > 0.5:
                        child_genes.append(parents[0][gene])
                        child_sigma.append(parents[0]['sigma'][g])
                    else:
                        child_genes.append(parents[1][gene])
                        child_sigma.append(parents[1]['sigma'][g])
                already_exists = (self.pop.loc[:, self.gene_names] == child_genes).all(1).any()
                if not already_exists:
                    break
                count += 1
            childs.append((child_genes, child_sigma))
        return childs

    def plot_genes(self, pop_member):
        # Iterate over all genes of the member
        for gene in pop_member.columns.tolist():
            print(f'{gene}:', end=" ")
            data = pop_member[gene].array[0]
            if isinstance(data, list):
                print('[', end="")
                for val in data:
                    print(f'{val}', end=", ")
                print(']', end=" ")
            else:
                print(float(np.round(data, decimals=5)), end=" ")

            # Print borders if available in initial gene pool
            try:
                min = np.round(self.initial_gene_pool[gene]['min'], decimals=2)
                max = np.round(self.initial_gene_pool[gene]['max'], decimals=2)
                print(f' [min: {min}, max: {max}]')
            except KeyError:
                print("")

    def pop_to_grid(self):
        return self.pop.loc[:, self.gene_names]

    def eval_fitness(self, target: list, *argv, **kwargs):
        raise NotImplementedError


class CGSGeneticAlgorithmTemplate(GeneticAlgorithmTemplate):
    def __init__(self, nodes, compute_dir=None, verbose: Optional[bool] = True):
        super().__init__()

        self.cgs = ClusterGridSearch(nodes=nodes, compute_dir=compute_dir, verbose=verbose)

    def eval_fitness(self, target: list, *argv, **kwargs):
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
# external imports
import re
from copy import deepcopy
from sys import intern
from typing import Union

# pyrates internal imports
from pyrates import PyRatesException
from pyrates.frontend.template.abc import AbstractBaseTemplate
from pyrates.frontend.utility import deep_freeze

# meta infos
from pyrates.ir.operator import OperatorIR

__author__ = " Daniel Rose"
__status__ = "Development"


class OperatorTemplate(AbstractBaseTemplate):
    """Generic template for an operator with a name, equations, variables and possible
    initialization conditions. The template can be used to create variations of a specific
    equations or variables."""

    cache = {}  # tracks all unique instances of applied operator templates
    target_ir = OperatorIR
    key_map = {}
    key_counter = {}

    # key_map = {}  # tracks renaming of of operator keys to have shorter unique keys

    def __init__(self, name: str, path: str, equations: Union[list, str], variables: dict,
                 description: str = "An operator template."):
        """For now: only allow single equations in operator template."""

        super().__init__(name, path, description)

        if isinstance(equations, str):
            self.equations = [intern(equations)]
        else:
            self.equations = [intern(eq) for eq in equations]
        self.variables = variables

    def update_template(self, name: str, path: str, equations: Union[str, list, dict] = None,
                        variables: dict = None, description: str = None):
        """Update all entries of the Operator template in their respective ways."""

        if equations:
            # if it is a string, just replace
            if isinstance(equations, str):
                equations = [equations]
            elif isinstance(equations, list):
                pass  # pass equations string to constructor
            # else, update according to predefined rules, assuming dict structure
            elif isinstance(equations, dict):
                equations = [_update_equation(eq, **equations) for eq in self.equations]

            else:
                raise TypeError("Unknown data type for attribute 'equations'.")
        else:
            # copy equations from parent template
            equations = self.equations

        if variables:
            variables = _update_variables(self.variables, variables)
        else:
            variables = self.variables

        rogue_variables = set()
        for var in variables:
            # remove variables that are not present in the equations
            found = False
            for eq in equations:
                if var in eq:
                    found = True

            if not found:
                # save entries in list, since dictionary must not change size during iteration
                rogue_variables.add(var)

        for var in rogue_variables:
            variables.pop(var)

        if not description:
            description = self.__doc__  # or do we want to enforce documenting a template?

        return self.__class__(name=name, path=path, equations=equations, variables=variables,
                              description=description)

    def apply(self, return_key=False, values: dict = None):
        """Returns the non-editable but unique, cashed definition of the operator."""

        # key for global operator cache is template name.
        # ToDo: Consider replacing this cache with a separate by-circuit cache.
        key = self.name
        if values is None:
            values = {}

        try:
            instance, default_values = self.cache[key]

            for vname, value in default_values.items():
                if vname not in values:
                    values[vname] = value

        except KeyError:
            # get variable definitions and specified default values
            variables = []
            default_values = dict()
            inputs = []
            output = None
            for vname, vtype, dtype, shape, value in _separate_variables(self.variables):

                # if no new value is given for a variable, get the default
                if vname not in values:
                    values[vname] = value

                default_values[vname] = value

                if vtype == "input":
                    inputs.append(vname)  # = dict(sources=[], reduce_dim=True)  # default to True for now
                    vtype = "state_var"
                elif vtype == "output":
                    if output is None:
                        output = vname  # for now assume maximum one output is present
                    else:
                        raise PyRatesException("More than one output specification found in operator. "
                                               "Only one output per operator is supported.")
                    vtype = "state_var"
                # pack variable defining properties into tuple
                variables.append((vname, vtype, dtype, shape))

            # reduce order of ODE if necessary
            # this step is currently skipped to streamline equation interface
            # instead equations need to be given as either non-differential equations or first-order
            # linear differential equations
            # *equations, variables = cls._reduce_ode_order(template.equations, variables)
            equations = self.equations

            instance = self.target_ir(equations=equations, variables=variables,
                                      inputs=inputs, output=output,
                                      template=self)
            self.cache[key] = (instance, default_values)

        if return_key:
            return instance, values, key
        else:
            return instance, values


def _update_variables(variables: dict, updates: dict):
    updated = deepcopy(variables)

    for var, var_dict in updates.items():
        if var in updated:
            # update dictionary defining single variable
            updated[var].update(var_dict)
        else:
            # copy new variable into variables dictionary
            updated.update({var: var_dict})

    return updated


def _update_equation(equation: str,  # original equation
                     replace: dict = False,  # replace parts of the string
                     remove: Union[list, tuple] = False,  # remove parts of the string
                     append: str = False,  # append to the end of the string
                     prepend: str = False,  # add to beginning of string
                     ):
    # replace existing terms by new ones
    if replace:
        for old, new in replace.items():
            equation = intern(equation.replace(old, new))
            # this might fail, if multiple replacements refer or contain the same variables
            # is it possible to call str.replace with tuples?

    # remove terms
    if remove:
        if isinstance(remove, str):
            equation = intern(equation.replace(remove, ""))
        else:
            for old in remove:
                equation = intern(equation.replace(old, ""))

    # append terms at the end of the equation string
    if append:
        # only allowing single append per update
        equation = intern(f"{equation} {append}")

        # prepend terms at the beginning of the equation string
    if prepend:
        # only allowing single prepend per update
        equation = intern(f"{prepend} {equation}")

    return equation


def _parse_defaults(expr: Union[str, int, float]):
    """Naive version of a parser for the default key of variables in a template. Returns data type,
    variable type and default value of the variable."""

    value = 0.  # Setting initial conditions for undefined variables to 0. Is that reasonable?
    if isinstance(expr, int):
        vtype = "constant"
        value = expr
        # dtype = "int32"
        dtype = "float32"  # default to float for maximum compatibility
    elif isinstance(expr, float):
        vtype = "constant"
        value = expr
        dtype = "float32"
        # restriction to 32bit float for consistency. May not be reasonable at all times.
    else:
        # set vtype
        if expr.startswith("input"):
            vtype = "input"
        elif expr.startswith("output"):
            vtype = "output"
        elif expr.startswith("variable"):
            vtype = "state_var"
        elif expr.startswith("constant"):
            vtype = "constant"
        elif expr.startswith("placeholder"):
            vtype = "placeholder"
        else:
            try:
                # if "." in expr:
                value = float(expr)  # default to float
                # else:
                #     value = int(expr)
                vtype = "constant"
            except ValueError:
                raise ValueError(f"Unable to interpret variable type in default definition {expr}.")

        # set dtype and value
        if expr.endswith("(float)"):
            dtype = "float32"  # why float32 and not float64?
        elif expr.endswith("(int)"):
            # dtype = "int32"
            dtype = "float32"  # default to float for maximum compatibility
        elif "." in expr:
            dtype = "float32"
            value = float(re.search("[+-]?([0-9]+([.][0-9]*)?|[.][0-9]+)", expr).group())
            # see https://stackoverflow.com/questions/12643009/regular-expression-for-floating-point-numbers
        elif re.search("[0-9]+", expr):
            # dtype = "int32"
            dtype = "float32"  # default to float for maximum compatibility
            value = float(re.search("[0-9]+", expr).group())
        else:
            dtype = "float32"  # base assumption

    return vtype, dtype, value


def _separate_variables(variables: dict):
    """
    Return variable definitions and the respective values.

    Returns
    -------
    variables
    inputs
    output
    """
    # this part can be improved a lot with a proper expression parser

    for vname, properties in variables.items():

        # get shape parameter, default shape is scalar
        shape = properties.get("shape", (1,))

        # identify variable type and data type
        # note: this assume that a "default" must be given for every variable
        try:
            vtype, dtype, value = _parse_defaults(properties["default"])
        except KeyError:
            raise PyRatesException("Variables need to have a 'default' (variable type, data type and/or value) "
                                   "specified.")

        yield vname, vtype, dtype, shape, value




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from typing import List

from pyrates import PyRatesException
from pyrates.ir.node import NodeIR
from pyrates.ir.operator import OperatorIR

__author__ = "Daniel Rose"
__status__ = "Development"


class EdgeIR(NodeIR):

    def __init__(self, operators: dict = None, values: dict = None, template: str = None):

        if not operators:
            # treat the special case of an empty operator graph
            operators = {}
            values = {}
            # operators = dict(identity_operator=OperatorIR(equations=["out_var = in_var"],
            #                                               variables=[("in_var", "state_var", "float32", (1,)),
            #                                                          ("out_var", "state_var", "float32", (1,))],
            #                                               inputs=["in_var"],
            #                                               output="out_var"))
            # values = dict(identity_operator=dict(in_var=0.,
            #                                      out_var=0.))

        super().__init__(operators, values, template)

    @property
    def input(self):
        """Detect input variable of edge, assuming only one input variable exists. This also references the operator
        the variable belongs to."""
        # noinspection PyTypeChecker
        in_op = [op for op, in_degree in self.op_graph.in_degree if in_degree == 0]  # type: List[str]

        # multiple input operations are possible, as long as they require the same singular input variable
        in_var = set()
        for op_key in in_op:
            for var in self[op_key].inputs:
                in_var.add(f"{op_key}/{var}")

        if len(in_var) == 1:
            return in_var.pop()
        elif len(in_var) == 0:
            return None
        else:
            raise PyRatesException("Too many input variables found. Exactly one or zero input variables are "
                                   "required per edge.")

    @property
    def input_var(self):
        return self.input

    @property
    def output(self):
        """Detect output variable of edge, assuming only one output variable exists."""

        # try to find single output variable
        # noinspection PyTypeChecker
        out_op = [op for op, out_degree in self.op_graph.out_degree if out_degree == 0]  # type: List[str]

        # only one single output operator allowed
        if len(out_op) == 1:
            out_var = self[out_op[0]].output
            return f"{out_op[0]}/{out_var}"
        elif len(out_op) == 0:
            return None
        else:
            raise PyRatesException("Too many or too little output operators found. Exactly one output operator and "
                                   "associated output variable is required per edge.")

    @property
    def output_var(self):
        return self.output

    def __hash__(self):
        raise NotImplementedError




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Contains functions that may be used as backend operations

"""

# external imports
import numpy as np

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


# function definitions
######################

def pr_sigmoid(x, scaling=1.0, steepness=1.0, offset=0.0):
    return scaling/(1. + np.exp(steepness*(offset-x)))


def pr_softmax(x, axis=0):
    x_exp = np.exp(x)
    return x_exp/np.sum(x_exp, axis=axis)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Some utility functions for parsing YAML-based definitions of circuits and components.
"""
from pyrates.frontend._registry import register_interface

__author__ = "Daniel Rose"
__status__ = "Development"


@register_interface
def to_dict(path: str):
    """Load a template from YAML and return the resulting dictionary.

    Parameters
    ----------

    path
        string containing path of YAML template of the form path.to.template or path/to/template.file.TemplateName.
        The dot notation refers to a path that can be found using python's import functionality. The slash notation
        refers to a file in an absolute or relative path from the current working directory.
    """
    from pyrates.frontend.file import parse_path

    template_name, filename, directory = parse_path(path)

    # test if file can be found (and potentially add extension)
    import os

    if "." in filename:
        filepath = os.path.join(directory, filename)
    else:
        # this is actually the default case for the internal interface
        for ext in ["yaml", "yml"]:
            filepath = os.path.join(directory, ".".join((filename, ext)))
            if os.path.exists(filepath):
                break
        else:
            raise FileNotFoundError(f"Could not identify file with name {filename} in directory {directory}.")

    # load as yaml file
    from ruamel.yaml import YAML

    yaml = YAML(typ="safe", pure=True)

    with open(filepath, "r") as file:
        file_dict = yaml.load(file)

    if template_name in file_dict:
        template_dict = file_dict[template_name]
        template_dict["path"] = path
        template_dict["name"] = template_name
    else:
        raise AttributeError(f"Could not find {template_name} in {filepath}.")

    return template_dict


@register_interface
def from_circuit(circuit, path: str, name: str):
    from pyrates.frontend.dict import from_circuit
    dict_repr = {name: from_circuit(circuit)}

    from ruamel.yaml import YAML
    yaml = YAML()

    from pyrates.utility.filestorage import create_directory
    create_directory(path)
    from pathlib import Path
    path = Path(path)
    yaml.dump(dict_repr, path)


@register_interface
def to_template(path: str, template_cls):
    return template_cls.from_yaml(path)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""This module provides the backend class that should be used to set up any backend in pyrates.
"""

# external imports
from typing import Optional, Tuple, List, Union, Any
from pandas import DataFrame, MultiIndex
import numpy as np
from networkx import find_cycle, NetworkXNoCycle, DiGraph
from copy import deepcopy
from warnings import filterwarnings

# pyrates imports
from pyrates.backend.numpy_backend import NumpyBackend as _NumpyBackend
from pyrates.backend.parser import parse_equation_system, parse_dict
from pyrates.ir.circuit import CircuitIR
from pyrates.frontend import CircuitTemplate
from pyrates.backend.parser import replace

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


class Compiler:
    """Helper class that contains all methods to compile a full circuit IR into its backend representation.

    Parameters
    ----------
    net_config
        Intermediate representation of the network configuration. For a more detailed description, see the documentation
        of `pyrates.IR.CircuitIR`.
    dt
        Step-size with which the network should be simulated later on.
        Important for discretizing delays, differential equations, ...
    vectorization
        Defines the mode of automatic parallelization optimization that should be used. Can be `nodes` for lumping all
        nodes together in a vector, `full` for full vectorization of the network, or `None` for no vectorization.
    name
        Name of the network.
    build_in_place
        If False, a copy of the `net_config``will be made before compute graph creation. Should be used, if the
        `net_config` will be re-used for multiple compute graphs.
    backend
        Backend in which to build the compute graph.
    solver
        Numerical solver to use for differential equations.

    """

    def __init__(self,
                 dt: float = 1e-3,
                 backend: Any = _NumpyBackend,
                 solver: str = 'euler',
                 float_precision: str = 'float32',
                 label: str = "circuit",
                 **kwargs
                 ) -> None:
        """Instantiates operator.
        """

        # filterwarnings("ignore", category=FutureWarning)

        # set basic attributes
        ######################

        self._float_precision = float_precision

        # instantiate the backend and set the backend default_device

        self.backend = backend
        self.solver = solver

        # pre-process the network configuration
        self.dt = dt
        # self._net_config_map = {}

        # set time constant of the network
        self.dt_parsed = parse_dict({'dt': {'vtype': 'constant', 'dtype': self._float_precision, 'shape': (),
                                             'value': self.dt}},
                                    backend=self.backend)['dt']

        # # move edge operations to nodes
        # ###############################
        #
        # print('building the compute graph...')
        #
        # # create equations and variables for each edge
        # for source_node, target_node, edge_idx in self.net_config.edges:
        #
        #     # extract edge information
        #     weight = self._get_edge_attr(source_node, target_node, edge_idx, 'weight')
        #     delay = self._get_edge_attr(source_node, target_node, edge_idx, 'delay')
        #     sidx = self._get_edge_attr(source_node, target_node, edge_idx, 'source_idx')
        #     tidx = self._get_edge_attr(source_node, target_node, edge_idx, 'target_idx')
        #     sop, svar, sval = self._get_edge_attr(source_node, target_node, edge_idx, 'source_var')
        #     top, tvar, tval = self._get_edge_attr(source_node, target_node, edge_idx, 'target_var')
        #     add_project = self._get_edge_attr(source_node, target_node, edge_idx, 'add_project')
        #     target_graph = self._get_node_attr(target_node, 'op_graph')
        #
        #     # define target index
        #     if delay is not None and tidx:
        #         tidx_tmp = []
        #         for idx, d in zip(tidx, delay):
        #             if type(idx) is list:
        #                 tidx_tmp.append(idx + [d])
        #             else:
        #                 tidx_tmp.append([idx, d])
        #         tidx = tidx_tmp
        #     elif not tidx and delay is not None:
        #         tidx = list(delay)
        #
        #     # create mapping equation and its arguments
        #     d = "[target_idx]" if tidx else ""
        #     idx = "[source_idx]" if sidx else ""
        #     assign = '+=' if add_project else '='
        #     eq = f"{tvar}{d} {assign} {svar}{idx} * weight"
        #     args = {}
        #     dtype = sval['dtype']
        #     args['weight'] = {'vtype': 'constant', 'dtype': dtype, 'value': weight}
        #     if tidx:
        #         args['target_idx'] = {'vtype': 'constant', 'dtype': 'int32',
        #                               'value': np.array(tidx, dtype=np.int32)}
        #     if sidx:
        #         args['source_idx'] = {'vtype': 'constant', 'dtype': 'int32',
        #                               'value': np.array(sidx, dtype=np.int32)}
        #     args[tvar] = tval
        #
        #     # add edge operator to target node
        #     op_name = f'edge_from_{source_node}_{edge_idx}'
        #     target_graph.add_node(op_name,
        #                           operator={'inputs': {svar: {'sources': [sop],
        #                                                       'reduce_dim': True,
        #                                                       'node': source_node}},
        #                                     'output': tvar,
        #                                     'equations': [eq]},
        #                           variables=args)
        #
        #     # connect edge operator to target operator
        #     target_graph.add_edge(op_name, top)
        #
        #     # add input information to target operator
        #     inputs = self._get_op_attr(target_node, top, 'inputs')
        #     if tvar in inputs.keys():
        #         inputs[tvar]['sources'].append(op_name)
        #     else:
        #         inputs[tvar] = {'sources': [op_name],
        #                         'reduce_dim': True}
        #
        # # collect node and edge operators
        # #################################
        #
        # variables = {'all/all/dt': self.dt_parsed}
        #
        # # edge operators
        # equations, variables_tmp = self._collect_op_layers(layers=[0], exclude=False, op_identifier="edge_from_")
        # variables.update(variables_tmp)
        # if equations:
        #     self.backend._input_layer_added = True
        #
        # # node operators
        # equations_tmp, variables_tmp = self._collect_op_layers(layers=[], exclude=True, op_identifier="edge_from_")
        # variables.update(variables_tmp)
        #
        # # bring equations into correct order
        # equations = sort_equations(edge_eqs=equations, node_eqs=equations_tmp)
        #
        # # parse all equations and variables into the backend
        # ####################################################
        #
        # self.backend.bottom_layer()
        #
        # # parse mapping
        # variables = parse_equation_system(equations=equations, equation_args=variables, backend=self.backend,
        #                                   solver=self.solver)
        #
        # # save parsed variables in net config
        # for key, val in variables.items():
        #     node, op, var = key.split('/')
        #     if "inputs" not in var and var != "dt":
        #         self._set_node_attr(node, var, val, op=op)

    def compile(self, circuit):
        """Transform CiruitIR instance into a form suitable for computation."""

        pass

    def run(self,
            simulation_time: Optional[float] = None,
            inputs: Optional[dict] = None,
            outputs: Optional[dict] = None,
            sampling_step_size: Optional[float] = None,
            out_dir: Optional[str] = None,
            verbose: bool = True,
            profile: Optional[str] = None,
            **kwargs
            ) -> Union[DataFrame, Tuple[DataFrame, float, float]]:
        """Simulate the backend behavior over time via a tensorflow session.

        Parameters
        ----------
        simulation_time
            Simulation time in seconds.
        inputs
            Inputs for placeholder variables. Each key is a tuple that specifies a placeholder variable in the graph
            in the following format: (node_name, op_name, var_name). Each value is an array that defines the input for
            the placeholder variable over time (first dimension).
        outputs
            Output variables that will be returned. Each key is the desired name of an output variable and each value is
            a tuple that specifies a variable in the graph in the following format: (node_name, op_name, var_name).
        sampling_step_size
            Time in seconds between sampling points of the output variables.
        out_dir
            Directory in which to store outputs.
        verbose
            If true, status updates will be printed to the console.
        profile
            Can be used to extract information about graph execution time and memory load. Can be:
            - `t` for returning the total graph execution time.
            - `m` for returning the peak memory consumption during graph excecution.
            - `mt` or `tm` for both

        Returns
        -------
        Union[DataFrame, Tuple[DataFrame, float, float]]
            First entry of the tuple contains the output variables in a pandas dataframe, the second contains the
            simulation time in seconds and the third the peak memory consumption. If profiling was not chosen during
            call of the function, only the dataframe will be returned.

        """

        filterwarnings("ignore", category=FutureWarning)

        # prepare simulation
        ####################

        if verbose:
            print("Preparing the simulation...")

        if not self._first_run:
            self.backend.remove_layer(0)
            self.backend.remove_layer(self.backend.top_layer())
        else:
            self._first_run = False

        # basic simulation parameters initialization
        if not simulation_time:
            simulation_time = self.dt
        sim_steps = int(simulation_time / self.dt)

        if not sampling_step_size:
            sampling_step_size = self.dt
        sampling_steps = int(sampling_step_size / self.dt)

        # add output variables to the backend
        #####################################

        # define output variables
        output_col = {}
        output_cols = []
        output_keys = []
        output_shapes = []
        if outputs:
            for key, val in outputs.items():
                val_split = val.split('/')
                node, op, var = "/".join(val_split[:-2]), val_split[-2], val_split[-1]
                for var_key, var_val in self.get_var(node=node, op=op, var=var, var_name=f"{key}_col").items():
                    var_shape = tuple(var_val.shape)
                    if var_shape in output_shapes:
                        idx = output_shapes.index(var_shape)
                        output_cols[idx].append(var_val)
                        output_keys[idx].append(var_key)
                    else:
                        output_cols.append([var_val])
                        output_keys.append([var_key])
                        output_shapes.append(var_shape)

                # create counting index for collector variables
                output_col.update(self.backend.add_output_layer(outputs=output_cols,
                                                                sampling_steps=int(sim_steps / sampling_steps),
                                                                out_shapes=output_shapes))

        # add input variables to the backend
        ####################################

        if inputs:

            inp_dict = dict()

            # linearize input dictionary
            for key, val in inputs.items():

                key_split = key.split('/')
                node, op, attr = "/".join(key_split[:-2]), key_split[-2], key_split[-1]

                if '_combined' in list(self.net_config.nodes.keys())[0]:

                    # fully vectorized case: add vectorized placeholder variable to input dictionary
                    var = self._get_node_attr(node=list(self.net_config.nodes.keys())[0], op=op, attr=attr)
                    inp_dict[var.name] = np.reshape(val, (sim_steps,) + tuple(var.shape))

                elif any(['_all' in key_tmp for key_tmp in self.net_config.nodes.keys()]):

                    # node-vectorized case
                    if node == 'all':

                        # go through all nodes, extract the variable and add it to input dict
                        i = 0
                        for node_tmp in self.net_config.nodes:
                            var = self._get_node_attr(node=node_tmp, op=op, attr=attr)
                            i_new = var.shape[0] if len(var.shape) > 0 else 1
                            inp_dict[var.name] = np.reshape(val[:, i:i_new], (sim_steps,) + tuple(var.shape))
                            i += i_new

                    elif node in self.net_config.nodes.keys():

                        # add placeholder variable of node(s) to input dictionary
                        var = self._get_node_attr(node=node, op=op, attr=attr)
                        inp_dict[var.name] = np.reshape(val, (sim_steps,) + tuple(var.shape))

                    elif any([node in key_tmp for key_tmp in self.net_config.nodes.keys()]) or \
                            any([node.split('.')[0] in key_tmp for key_tmp in self.net_config.nodes.keys()]):

                        node_tmp = node.split('.')[0] if '.' in node else node

                        # add vectorized placeholder variable of specified node type to input dictionary
                        for node_tmp2 in list(self.net_config.nodes.keys()):
                            if node_tmp in node_tmp2:
                                break
                        var = self._get_node_attr(node=node_tmp2, op=op, attr=attr)
                        inp_dict[var.name] = np.reshape(val, (sim_steps,) + tuple(var.shape))

                else:

                    # non-vectorized case
                    if node == 'all':

                        # go through all nodes, extract the variable and add it to input dict
                        for i, node_tmp in enumerate(self.net_config.nodes.keys()):
                            var = self._get_node_attr(node=node_tmp, op=op, attr=attr)
                            inp_dict[var.name] = np.reshape(val[:, i], (sim_steps,) + tuple(var.shape))

                    elif any([node in key_tmp for key_tmp in self.net_config.nodes.keys()]):

                        # extract variables from nodes of specified type
                        i = 0
                        for node_tmp in self.net_config.nodes.keys():
                            if node in node_tmp:
                                var = self._get_node_attr(node=node_tmp, op=op, attr=attr)
                                inp_dict[var.name] = np.reshape(val[:, i], (sim_steps,) + tuple(var.shape))
                                i += 1

            self.backend.add_input_layer(inputs=inp_dict)

        # run simulation
        ################

        if verbose:
            print("Running the simulation...")

        if profile is None:
            output_col = self.backend.run(steps=sim_steps, outputs=output_col, sampling_steps=sampling_steps,
                                          out_dir=out_dir, profile=profile, **kwargs)
        else:
            output_col, time, memory = self.backend.run(steps=sim_steps, outputs=output_col, out_dir=out_dir,
                                                        profile=profile, sampling_steps=sampling_steps, **kwargs)

        if verbose and profile:
            if simulation_time:
                print(f"{simulation_time}s of backend behavior were simulated in {time} s given a "
                      f"simulation resolution of {self.dt} s.")
            else:
                print(f"ComputeGraph computations finished after {time} seconds.")
        elif verbose:
            print('finished!')

        # store output variables in data frame
        ######################################

        # ungroup grouped output variables
        outputs = {}
        for names, group_key in zip(output_keys, output_col.keys()):
            out_group = output_col[group_key]
            for i, key in enumerate(names):
                outputs[key] = out_group[:, i]

        out_var_vals = []
        out_var_names = []
        for key in list(outputs):

            var = outputs.pop(key)
            if len(var.shape) > 1 and var.shape[1] > 1:
                for i in range(var.shape[1]):
                    var_tmp = var[:, i]
                    if len(var.shape) > 1:
                        var_tmp = np.squeeze(var_tmp)
                    out_var_vals.append(var_tmp)
                    key_split = key.split('/')
                    var_name = key_split[-1]
                    var_name = var_name[:var_name.find('_col')]
                    node_name = "/".join(key_split[:-1])
                    out_var_names.append((var_name, f'{node_name}_{i}'))
            else:
                if len(var.shape) > 1:
                    var = np.squeeze(var)
                out_var_vals.append(var)
                key_split = key.split('/')
                var_name = key_split[-1]
                var_name = var_name[:var_name.find('_col')]
                node_name = "/".join(key_split[:-1])
                out_var_names.append((var_name, node_name))

        # create multi-index
        index = MultiIndex.from_tuples(out_var_names, names=['var', 'node'])

        # create dataframe
        if out_var_vals:
            data = np.asarray(out_var_vals).T
            if len(data.shape) > 2:
                data = data.squeeze()
            idx = np.arange(0., simulation_time, sampling_step_size)[-data.shape[0]:]
            out_vars = DataFrame(data=data[0:len(idx), :],
                                 index=idx,
                                 columns=index)
        else:
            out_vars = DataFrame()

        # return results
        ################

        if profile:
            return out_vars, time, memory
        return out_vars

    def get_var(self, node: str, op: str, var: str, var_name: Optional[str] = None, **kwargs) -> dict:
        """Extracts a variable from the graph.

        Parameters
        ----------
        node
            Name of the node(s), the variable exists on. Can be 'all' for all nodes, or a sub-string that defines a
            class of nodes or a specific node name.
        op
            Name of the operator the variable belongs to.
        var
            Name of the variable.
        var_name
            Name under which the variable should be returned
        kwargs
            Additional keyword arguments that may be used to pass arguments for the backend like name scopes.

        Returns
        -------
        dict
            Dictionary with all variables found in the network that match the provided signature.

        """

        if not var_name:
            var_name = var
        var_col = {}

        if node == 'all':

            # collect output variable from every node in backend
            for node in self.net_config.nodes.keys():
                var_col[f'{node}/{op}/{var_name}'] = self._get_node_attr(node=node, op=op, attr=var)

        elif node in self.net_config.nodes.keys() or node in self._net_config_map.keys():

            # get output variable of specific backend node
            var_col[f'{node}/{op}/{var_name}'] = self._get_node_attr(node=node, op=op, attr=var, **kwargs)

        elif any([node in key for key in self.net_config.nodes.keys()]):

            # get output variable from backend nodes of a certain type
            for node_tmp in self.net_config.nodes.keys():
                if node in node_tmp:
                    var_col[f'{node}/{op}/{var_name}'] = self._get_node_attr(node=node_tmp, op=op, attr=var, **kwargs)
        else:

            # get output variable of specific, vectorized backend node
            i = 0
            for node_tmp in self._net_config_map.keys():
                if node in node_tmp:
                    var_col[f'{node}/{op}/{var_name}_{i}'] = self._get_node_attr(node=node_tmp, op=op, attr=var,
                                                                                 **kwargs)
                    i += 1

        return var_col

    def clear(self):
        """Clears the backend graph from all operations and variables.
        """
        self.backend.clear()

    def _net_config_consistency_check(self, circuit: CircuitIR) -> CircuitIR:
        """Checks whether the passed network configuration follows the expected intermediate representation structure.

        Parameters
        ----------
        circuit
            Intermediate representation of the network configuration that should be translated into the backend.

        Returns
        -------
        CircuitIR
            The checked (and maybe slightly altered) network configuration.

        """

        # check node attributes
        #######################

        # go through each node in the network  config
        for node_name, data in circuit.nodes(data=True):
            node = data["node"]
            # check whether an operation graph exists
            try:
                op_graph = node.op_graph
            except KeyError:
                raise KeyError(f'Key `node` not found on node {node_name}. Every node in the network configuration '
                               f'needs a field with the key `node` under which '
                               f'its operation graph and its template is stored.')
            except AttributeError:
                raise AttributeError(f'Attribute `op_graph` not found on node {node_name}. Every node in the network '
                                     f'configuration needs a graph object stored under its `node` key, which contains '
                                     f'all information regarding its operators, variables,'
                                     f'and input-output relationships.')

            # go through the operations on the node
            for op_name, data in op_graph.nodes(data=True):
                op = data["operator"]
                # check whether the variable field exists on the operator
                try:
                    variables = op.variables
                except KeyError:
                    raise KeyError(f'Key `variables` not found in operator {op_name} of node {node_name}. Every '
                                   f'operator on a node needs a field `variables` under which all necessary '
                                   'variables are defined via key-value pairs.')

                # check whether the operator field exists on the operator
                # try:
                #     op_info = op['operator']
                # except KeyError:
                #     raise KeyError(f'Key `operator` not found in operator {op_name} of node {node_name}. Every '
                #                    f'operator on a node needs a field `operator` under which all the equations, '
                #                    'inputs and output are defined.')

                # go through the variables
                for var_name, var in variables.items():

                    # check definition of each variable
                    var_def = {'state_var': {'value': np.zeros((1,), dtype=np.float32),
                                             'dtype': self._float_precision,
                                             'shape': (1,),
                                             },
                               'constant': {'value': KeyError,
                                            'shape': (1,),
                                            'dtype': self._float_precision},
                               'placeholder': {'dtype': self._float_precision,
                                               'shape': (1,),
                                               'value': None},
                               'raw': {'dtype': None,
                                       'shape': None,
                                       'value': KeyError}
                               }
                    try:
                        vtype = var.vtype
                    except KeyError:
                        raise KeyError(f'Field `vtype` not found in variable definition of variable {var_name} of '
                                       f'operator {op_name} on node {node_name}. Each variable needs a field `vtype`'
                                       f'that indicates whether the variable is a `state_var`, `constant` or '
                                       f'`placeholder`.')
                    var_defs = var_def[vtype]

                    # check the value of the variable
                    # try:
                    #     _ = node.values[op_name][var_name]
                    # except KeyError:
                    #     if var_defs['value'] is KeyError:
                    #         raise KeyError(f'Field `value` not found in variable definition of variable {var_name} of '
                    #                        f'operator {op_name} on node {node_name}, but it is needed for variables of '
                    #                        f'type {vtype}.')
                    #     elif var_defs['value'] is not None:
                    #         var['value'] = var_defs['value']

                    # check the shape of the variable
                    try:
                        shape = var.shape
                        if len(shape) == 0:
                            var['shape'] = (1,)
                    except KeyError:
                        var['shape'] = var_defs['shape']

                    # check the data type of the variable
                    try:
                        _ = var.dtype
                    except KeyError:
                        var['dtype'] = var_defs['dtype']

                    val = node.values[op_name][var_name]
                    if not hasattr(val, "shape"):
                        node.values[op_name][var_name] = np.zeros(var.shape, dtype=var.dtype) + val

                # check whether the equations, inputs and output fields exist on the operator field
                op_fields = ['equations', 'inputs', 'output']
                for field in op_fields:
                    try:
                        _ = getattr(op, field)
                    except KeyError:
                        if field == 'equations':
                            raise KeyError(f'Field {field} not found in `operators` field of operator {op_name} on '
                                           f'node {node_name}. Each operator should follow a list of equations that '
                                           f'needs to be provided at this position.')
                        elif field == 'inputs':
                            op['inputs'] = {}
                        else:
                            raise KeyError(f'Field {field} not found in `operators` field of operator {op_name} on '
                                           f'node {node_name}. Each operator should have an output, the name of which '
                                           f'needs to be provided at this position.')

        # check edge attributes
        #######################

        # go through edges
        for source, target, idx in circuit.edges:

            edge = circuit.edges[source, target, idx]

            # check whether source and target variable information were provided
            try:
                _ = edge['source_var']
            except KeyError:
                raise KeyError(f'Field `source_var` not found on edge {idx} between {source} and {target}. Every edge '
                               f'needs information about the variable on the source node it should access that needs '
                               f'to be provided at this position.')
            try:
                _ = edge['target_var']
            except KeyError:
                raise KeyError(f'Field `target_var` not found on edge {idx} between {source} and {target}. Every edge '
                               f'needs information about the variable on the target node it should access that needs '
                               f'to be provided at this position.')

            # check weight of edge
            try:
                weight = edge['weight']
                if not hasattr(weight, 'shape') or not weight.shape:
                    weight = np.ones((1,), dtype=self._float_precision) * weight
                elif 'float' not in str(weight.dtype):
                    weight = np.asarray(weight, dtype=self._float_precision)
                edge['weight'] = weight
            except KeyError:
                edge['weight'] = np.ones((1,), dtype=self._float_precision)

            # check delay of edge
            try:
                delay = edge['delay']
                if delay is not None:
                    if not hasattr(delay, 'shape') or not delay.shape:
                        delay = np.zeros((1,)) + delay
                    if 'float' in str(delay.dtype):
                        delay = np.asarray((delay / self.dt) + 1, dtype=np.int32)
                    edge['delay'] = delay
            except KeyError:
                edge['delay'] = None

        return circuit

    def _collect_ops(self, ops: List[str], node_name: str) -> tuple:
        """Adds a number of operations to the backend graph.

        Parameters
        ----------
        ops
            Names of the operators that should be parsed into the graph.
        node_name
            Name of the node that the operators belong to.

        Returns
        -------
        tuple
            Collected and updated operator equations and variables

        """

        # set up update operation collector variable
        equations = []
        variables = {}

        # add operations of same hierarchical lvl to compute graph
        ############################################################

        for op_name in ops:

            # retrieve operator and operator args
            op_args = self._get_op_attr(node_name, op_name, 'variables')
            op_args['inputs'] = {}
            op_info = self._get_op_attr(node_name, op_name, 'operator')

            if getattr(op_info, 'collected', False):
                break

            # handle operator inputs
            in_ops = {}
            for var_name, inp in op_info['inputs'].items():

                # go through inputs to variable
                if inp['sources']:

                    in_ops_col = {}
                    reduce_inputs = inp['reduce_dim'] if type(inp['reduce_dim']) is bool else False
                    in_node = inp['node'] if 'node' in inp else node_name

                    for i, in_op in enumerate(inp['sources']):

                        # collect single input to op
                        in_var = self._get_op_attr(in_node, in_op, 'output', retrieve=False)
                        try:
                            in_val = self._get_op_attr(in_node, in_op, 'output', retrieve=True)
                        except KeyError:
                            in_val = None
                        in_ops_col[f"{in_node}/{in_op}/{in_var}"] = in_val

                    if len(in_ops_col) > 1:
                        in_ops[var_name] = self._map_multiple_inputs(in_ops_col, reduce_inputs)
                    else:
                        key, _ = in_ops_col.popitem()
                        in_node, in_op, in_var = key.split("/")
                        in_ops[var_name] = (in_var, {in_var: key})

            # replace input variables with input in operator equations
            for var, inp in in_ops.items():
                for i, eq in enumerate(op_info['equations']):
                    op_info['equations'][i] = replace(eq, var, inp[0], rhs_only=True)
                op_args['inputs'].update(inp[1])

            # collect operator variables and equations
            scope = f"{node_name}/{op_name}"
            variables[f"{scope}/inputs"] = {}
            equations += [(eq, scope) for eq in op_info['equations']]
            for key, var in op_args.items():
                full_key = f"{scope}/{key}"
                if key == 'inputs':
                    variables[f"{scope}/inputs"].update(var)
                elif full_key not in variables:
                    variables[full_key] = var
            try:
                setattr(op_info, 'collected', True)
            except AttributeError:
                op_info['collected'] = True

        return equations, variables

    # def _get_node_attr(self, node: str, attr: str, op: Optional[str] = None, **kwargs) -> Any:
    #     """Extract attribute from node of the network.
    #
    #     Parameters
    #     ----------
    #     node
    #         Name of the node.
    #     attr
    #         Name of the attribute on the node.
    #     op
    #         Name of the operator. Only needs to be provided for operator variables.
    #
    #     Returns
    #     -------
    #     Any
    #         Node attribute.
    #
    #     """
    #
    #     if op:
    #         return self._get_op_attr(node, op, attr, **kwargs)
    #     try:
    #         return self.net_config[node][attr]
    #     except KeyError:
    #         vals = []
    #         for op in self.net_config[node]:
    #             vals.append(self._get_op_attr(node, op, attr, **kwargs))
    #         return vals

    # def _set_node_attr(self, node: str, attr: str, val: Any, op: Optional[str] = None) -> Any:
    #     """Sets attribute on a node.
    #
    #     Parameters
    #     ----------
    #     node
    #         Name of the node.
    #     attr
    #         Name of the attribute.
    #     op
    #         Name of the operator, the attribute belongs to. Does not need to be provided for attributes that do not
    #         belong to an operator.
    #
    #     Returns
    #     -------
    #     Any
    #         New node attribute value.
    #
    #     """
    #
    #     if op:
    #         return self._set_op_attr(node, op, attr, val)
    #
    #     try:
    #         self.net_config.nodes[node]['node'][attr] = val
    #     except KeyError:
    #         vals_updated = []
    #         for op in self.net_config.nodes[node]['node']['op_graph'].nodes:
    #             vals_updated.append(self._set_op_attr(node, op, attr, val))
    #         return vals_updated

    # def _get_edge_attr(self, source: str, target: str, edge: Union[str, int], attr: str,
    #                    retrieve_from_node: bool = True) -> Any:
    #     """Extracts attribute from edge.
    #
    #     Parameters
    #     ----------
    #     source
    #         Name of the source node.
    #     target
    #         Name of the target node.
    #     edge
    #         Index or name of the edge.
    #     attr
    #         Name of the attribute
    #     retrieve_from_node
    #         If true, the attribute will be retrieved from the source or target node. Only relevant if the attribute is
    #         'source_var' or 'target_var'. Else, the value of the attribute on the edge will be returned.
    #
    #     Returns
    #     -------
    #     Any
    #         Edge attribute.
    #
    #     """
    #
    #     try:
    #         attr_val = self.net_config.edges[source, target, edge][attr]
    #         if 'var' in attr and type(attr_val) is str and retrieve_from_node:
    #             op, var = attr_val.split('/')
    #             if 'source' in attr:
    #                 attr_val = self._get_op_attr(source, op, var)
    #             else:
    #                 attr_val = self._get_op_attr(target, op, var)
    #             return op, var, attr_val
    #     except KeyError:
    #         attr_val = None
    #
    #     return attr_val

    # def _set_edge_attr(self, source: str, target: str, edge: Union[int, str], attr: str, val: Any) -> Any:
    #     """Sets value of an edge attribute.
    #
    #     Parameters
    #     ----------
    #     source
    #         Name of the source node.
    #     target
    #         Name of the target node.
    #     edge
    #         Name or index of the edge.
    #     attr
    #         Name of the attribute.
    #     val
    #         New value of the attribute.
    #
    #     Returns
    #     -------
    #     Any
    #         Value of the updated edge attribute.
    #
    #     """
    #
    #     self.net_config.edges[source, target, edge][attr] = val
    #     return self.net_config.edges[source, target, edge][attr]

    # def _get_op_attr(self, node: str, op: str, attr: str, retrieve: bool = True, **kwargs) -> Any:
    #     """Extracts attribute of an operator.
    #
    #     Parameters
    #     ----------
    #     node
    #         Name of the node.
    #     op
    #         Name of the operator on the node.
    #     attr
    #         Name of the attribute of the operator on the node.
    #     retrieve
    #         If attribute is output, this can be set to True to receive the handle to the output variable, or to false
    #         to receive the name of the output variable.
    #
    #     Returns
    #     -------
    #     Any
    #         Operator attribute.
    #
    #     """
    #
    #     if node in self._net_config_map and op in self._net_config_map[node] and attr in self._net_config_map[node][op]:
    #         node, op, attr, attr_idx = self._net_config_map[node][op][attr]
    #         idx = f"{list(attr_idx)}" if type(attr_idx) is tuple else attr_idx
    #         return self.backend.apply_idx(self._get_op_attr(node, op, attr), idx) if retrieve else attr_idx
    #     elif node in self.net_config:
    #         op = self.net_config[node]['op_graph'].nodes[op]
    #     else:
    #         raise ValueError(f'Node with name {node} is not part of this network.')
    #
    #     if attr == 'output' and retrieve:
    #         attr = op['operator']['output']
    #
    #     if attr in op['variables'].keys():
    #         attr_val = op['variables'][attr]
    #     elif hasattr(op['operator'], 'keys') and attr in op['operator'].keys():
    #         attr_val = op['operator'][attr]
    #     elif hasattr(op['operator'], attr):
    #         attr_val = getattr(op['operator'], attr)
    #     else:
    #         try:
    #             attr_val = op[attr]
    #         except KeyError as e:
    #             try:
    #                 attr_val = getattr(op, attr)
    #             except AttributeError:
    #                 raise e
    #
    #     return attr_val

    # def _set_op_attr(self, node: str, op: str, attr: str, val: Any) -> Any:
    #     """Sets value of an operator attribute.
    #
    #     Parameters
    #     ----------
    #     node
    #         Name of the node.
    #     op
    #         Name of the operator on the node.
    #     attr
    #         Name of the operator attribute on the node.
    #     val
    #         New value of the operator attribute.
    #
    #     Returns
    #     -------
    #     Any
    #         Updated attribute value.
    #
    #     """
    #
    #     op = self.net_config[node]['op_graph'].nodes[op]
    #     if attr in op['variables'].keys():
    #         op['variables'][attr] = val
    #         return op['variables'][attr]
    #     elif attr == 'output':
    #         op['variables'][op[attr]] = val
    #         return op['variables'][op[attr]]
    #     elif hasattr(op['operator'], attr):
    #         setattr(op['operator'], attr, val)
    #         return getattr(op['operator'], attr)
    #     else:
    #         try:
    #             op[attr] = val
    #             return op[attr]
    #         except KeyError:
    #             return None

    def _collect_op_layers(self, circuit, layers: list, exclude: Optional[bool] = False, op_identifier: Optional[str] = None
                           ) -> tuple:
        """

        Parameters
        ----------
        layers
        exclude
        op_identifier

        Returns
        -------

        """

        equations = []
        variables = {}

        for node_name, data in circuit.nodes(data=True):

            node = data["node"]
            op_graph = node.op_graph
            graph = op_graph.copy()  # type: DiGraph

            # go through all operators on node and pre-process + extract equations and variables
            i = 0
            while graph.nodes:

                # get all operators that have no dependencies on other operators
                # noinspection PyTypeChecker
                ops = [op for op, in_degree in graph.in_degree if in_degree == 0]

                if (i in layers and not exclude) or (i not in layers and exclude):

                    if op_identifier:
                        ops_tmp = [op for op in ops if op_identifier not in op] if exclude else \
                            [op for op in ops if op_identifier in op]
                    else:
                        ops_tmp = ops
                    op_eqs, op_vars = self._collect_ops(ops_tmp, node_name=node_name)

                    # collect primary operator equations and variables
                    if i == len(equations):
                        equations.append(op_eqs)
                    else:
                        equations[i] += op_eqs
                    for key, var in op_vars.items():
                        if key not in variables:
                            variables[key] = var

                # remove parsed operators from graph
                graph.remove_nodes_from(ops)
                i += 1

        return equations, variables

    @staticmethod
    def _apply_idx(var: Any, idx: Optional[Union[int, tuple]] = None) -> Any:
        """Applies index to graph variable. Only used for observing graph variables,
        not for graph internal computations.

        Parameters
        ----------
        var
            Graph variable.
        idx
            Index to the variable dimensions.

        Returns
        -------
        Any
            Indexed variable.

        """

        if idx is None:
            return var
        elif type(idx) is tuple:
            return var[idx[0]:idx[1]]
        else:
            return var[idx]

    def _get_nodes_with_attr(self, attr: str, val: Any) -> list:
        """Extracts nodes from graph for which a certain attribute takes on a certain value.

        Parameters
        ----------
        attr
            Attribute of interest on the nodes.
        val
            Attribute value that selected nodes should have.

        Returns
        -------
        list
            Nodes for which the attribute takes on the indicated value.

        """

        nodes = []
        for node in self.net_config.nodes:
            test_val = self._get_node_attr(node, attr)
            if hasattr(val, 'nodes') and val.nodes.keys() == test_val.nodes.keys():
                nodes.append(node)
            elif val == test_val:
                nodes.append(node)

        return nodes

    def _contains_node(self, node: str, target_node: str) -> bool:
        """Checks whether a vectorized graph node contains an original, non-vectorized target node or not.

        Parameters
        ----------
        node
            Vectorized graph node.
        target_node
            Less vectorized node.

        Returns
        -------
        bool
            Indicates whether target node is contained in the vectorized node or not.

        """

        try:
            node_map = self._net_config_map[target_node]
            for op in node_map.values():
                for var in op.values():
                    if var[0] == node:
                        return True
            else:
                return False
        except KeyError:
            raise KeyError(f'Could not identify target node {target_node}.')

    def _get_edges_between(self, source: str, target: str) -> List[tuple]:
        """Extract all edges between a source and a target node.

        Parameters
        ----------
        source
            Name of the source node.
        target
            Name of the target node.

        Returns
        -------
        List[tuple]
            Collection of all edges between source and target.

        """

        if ('_all' in source and '_all' in target) or ('_combined' in source and '_combined' in target):
            edges = []
            for source_tmp in self._net_config_map:
                for target_tmp in self._net_config_map:
                    if self._contains_node(source, source_tmp) and self._contains_node(target, target_tmp):
                        edges += [(source_tmp, target_tmp, edge) for edge
                                  in range(self.net_config.graph.number_of_edges(source_tmp, target_tmp))]
            return edges
        else:
            return [(source, target, edge) for edge in range(self.net_config.graph.number_of_edges(source, target))]

    def _get_edge_conn(self, source: str, target: str, edge: Union[int, str]) -> Tuple[np.ndarray, np.ndarray]:
        """Extracts weight and delay vectors of an edge.

        Parameters
        ----------
        source
            Name of the source node.
        target
            Name of the target node.
        edge
            Name of index of the edge.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Weight and delay of the edge (both are arrays).

        """

        # check delay of edge
        delay = self._get_edge_attr(source, target, edge, 'delay')
        if delay is not None and delay.ndim > 1:
            raise ValueError(f"Automatic optimization of the graph (i.e. method `vectorize`"
                             f" cannot be applied to networks with variables of 2 or more"
                             f" dimensions. Delay of edge {edge} between {source} and {target} has shape"
                             f" {delay.shape}. Please turn of the `vectorize` option or change"
                             f" the edges' dimensionality.")

        # check weight of edge
        weight = self._get_edge_attr(source, target, edge, 'weight')
        if weight is not None and weight.ndim > 1:
            raise ValueError(f"Automatic optimization of the graph (i.e. method `vectorize`"
                             f" cannot be applied to networks with variables of 2 or more"
                             f" dimensions. Weight of edge {edge} between {source} and {target} has shape"
                             f" {weight.shape}. Please turn of the `vectorize` option or"
                             f" change the edges' dimensionality.")

        # match dimensionality of delay and weight
        if delay is None or weight is None:
            pass
        elif delay.shape[0] > 1 and weight.shape[0] == 1:
            weight = np.tile(weight, (delay.shape[0], 1))
        elif weight.shape[0] > 1 and delay.shape[0] == 1:
            delay = np.tile(delay, (weight.shape[0], 1))
        elif delay.shape != weight.shape:
            raise ValueError(f"Dimensionality of weights and delays of edge between {source} and {target}"
                             f" do not match. They are of shape {weight.shape} and {delay.shape}."
                             f" Please turn of the `vectorize` option or change the dimensionality of the"
                             f" edge's attributes.")

        return weight, delay

    def _get_edge_var_idx(self, source: str, target: str, edge: Union[int, str], idx_type: str) -> Union[list, None]:
        """Extracts indices of an edge that map vectorized weights and delays to the original edges.

        Parameters
        ----------
        source
            Name of the source node.
        target
            Name of the target node.
        edge
            Index or name of the edge.

        Returns
        -------
        Union[list, None]
            List of indices.

        """

        if idx_type == 'source':
            var = 'source_idx'
            node_var = 'source_var'
            node_to_idx = source
        elif idx_type == 'target':
            var = 'target_idx'
            node_var = 'target_var'
            node_to_idx = target
        else:
            raise ValueError('Wrong `idx_type`. Please, choose either `source` or `target`.')

        var_idx = self._get_edge_attr(source, target, edge, var)
        if var_idx:
            op, var = self.net_config.edges[source, target, edge][node_var].split('/')
            _, _, _, idx = self._net_config_map[node_to_idx][op][var]
            if type(idx) is tuple:
                idx = range(idx[0], idx[1])
                idx_new = []
                for i in var_idx:
                    if type(i) is tuple:
                        idx_new.append([idx[i[0]:i[1]]])
                    elif type(i) is list:
                        idx_new.append([idx[i[0]]])
                    else:
                        idx_new.append([idx[i]])
            else:
                idx_new = [[idx]]
        else:
            try:
                op, var = self.net_config.edges[source, target, edge][node_var].split('/')
                _, _, _, idx_new = self._net_config_map[node_to_idx][op][var]
                if type(idx_new) is int:
                    idx_new = [idx_new]
                else:
                    idx_new = list(idx_new)
            except KeyError:
                idx_new = None

        return idx_new

    @staticmethod
    def _sort_edges(circuit, edges: List[tuple], attr: str) -> dict:
        """Sorts edges according to the given edge attribute.

        Parameters
        ----------
        edges
            Collection of edges of interest.
        attr
            Name of the edge attribute.

        Returns
        -------
        dict
            Key-value pairs of the different values the attribute can take on (keys) and the list of edges for which
            the attribute takes on that value (value).

        """

        edges_new = {}
        edge_idx = {}
        for edge in edges:
            if len(edge) == 3:
                source, target, edge = edge
            else:
                source, target = edge
                if (source, target) not in edge_idx:
                    edge_idx[(source, target)] = 0
                edge = edge_idx[(source, target)]
                edge_idx[(source, target)] += 1
            value = circuit.edges[source, target, edge][attr]
            if value not in edges_new.keys():
                edges_new[value] = [(source, target, edge)]
            else:
                edges_new[value].append((source, target, edge))

        return edges_new

    def _add_edge_buffer(self, circuit, node, op: str, var: str, idx: int, buffer_length: int, edge: tuple) -> None:
        """Adds a buffer variable to an edge.

        Parameters
        ----------
        node
            Name of the target node of the edge.
        op
            Name of the target operator of the edge.
        var
            Name of the target variable of the edge.
        idx
            Index of the buffer variable for that specific edge.
        buffer_length
            Length of the time-buffer that should be added to realize edge delays.
        edge
            Edge identifier (source_name, target_name, edge_idx).

        Returns
        -------
        None

        """

        var_path = f"{op}/{var}"
        target_shape = node[var_path].shape
        op_graph = node.op_graph

        # create buffer variable definitions
        if len(target_shape) < 1 or (len(target_shape) == 1 and target_shape[0] == 1):
            buffer_shape = (buffer_length + 1,)
            buffer_shape_reset = (1,)
        else:
            buffer_shape = (target_shape[0], buffer_length + 1)
            buffer_shape_reset = (target_shape[0], 1)
        var_dict = {f'{var}_buffer_{idx}': {'vtype': 'state_var',
                                            'dtype': self._float_precision,
                                            'shape': buffer_shape,
                                            'value': 0.
                                            },
                    f'{var}_buffer_{idx}_reset': {'vtype': 'constant',
                                                  'dtype': self._float_precision,
                                                  'shape': buffer_shape_reset,
                                                  'value': 0.
                                                  }
                    }

        # create buffer equations
        if len(target_shape) < 1 or (len(target_shape) == 1 and target_shape[0] == 1):
            eqs_op_read = [f"{var} = {var}_buffer_{idx}[0]"]
            eqs_op_rotate = [f"{var}_buffer_{idx} = concat(({var}_buffer_{idx}[1:], {var}_buffer_{idx}_reset), 0)"]
        else:
            eqs_op_read = [f"{var} = {var}_buffer_{idx}[:, 0]"]
            eqs_op_rotate = [f"{var}_buffer_{idx} = concat(({var}_buffer_{idx}[:, 1:], {var}_buffer_{idx}_reset), 1)"]

        # add buffer operators to operator graph
        op_graph.add_node(f'{op}_{var}_buffer_rotate_{idx}',
                          operator={'inputs': {},
                                    'output': f'{var}_buffer_{idx}',
                                    'equations': eqs_op_rotate},
                          variables=var_dict)
        op_graph.add_node(f'{op}_{var}_buffer_read_{idx}',
                          operator={'inputs': {f'{var}_buffer_{idx}': {'sources': [f'{op}_{var}_buffer_rotate_{idx}'],
                                                                       'reduce_dim': False}},
                                    'output': var,
                                    'equations': eqs_op_read},
                          variables={})

        # connect operators to rest of the graph
        op_graph.add_edge(f'{op}_{var}_buffer_rotate_{idx}', f'{op}_{var}_buffer_read_{idx}')
        op_graph.add_edge(f'{op}_{var}_buffer_read_{idx}', op)

        # add input information to target operator
        inputs = op_graph.nodes[op]["inputs"]
        if var in inputs.keys():
            inputs[var]['sources'].add(f'{op}_{var}_buffer_read_{idx}')
        else:
            inputs[var] = {'sources': {f'{op}_{var}_buffer_read_{idx}'},
                           'reduce_dim': True}

        # update edge information
        edge_dict = circuit.edges[edge]
        edge_dict['target_var'] = f'{op}_{var}_buffer_rotate_{idx}/{var}_buffer_{idx}'
        edge_dict['add_project'] = True

    def _add_edge_input_collector(self, circuit, node, op: str, var: str, idx: int, edge: tuple) -> None:
        """Adds an input collector variable to an edge.

        Parameters
        ----------
        node
            Name of the target node of the edge.
        op
            Name of the target operator of the edge.
        var
            Name of the target variable of the edge.
        idx
            Index of the input collector variable on that edge.
        edge
            Edge identifier (source_name, target_name, edge_idx).

        Returns
        -------
        None

        """

        var_path = f"{op}/{var}"
        target_shape = node[var_path].shape
        op_graph = node.op_graph

        # create collector equation
        eqs = [f"{var} = {var}_col_{idx}"]

        # create collector variable definition
        var_dict = {f'{var}_col_{idx}': {'vtype': 'state_var',
                                         'dtype': self._float_precision,
                                         'shape': target_shape,
                                         'value': 0.
                                         }}

        # add collector operator to operator graph
        op_graph.add_node(f'{op}_{var}_col_{idx}',
                          operator={'inputs': {},
                                    'output': var,
                                    'equations': eqs},
                          variables=var_dict)
        op_graph.add_edge(f'{op}_{var}_col_{idx}', op)

        # add input information to target operator
        op_inputs = op_graph.nodes[op]["inputs"]
        if var in op_inputs.keys():
            op_inputs[var]['sources'].add(f'{op}_{var}_col_{idx}')
        else:
            op_inputs[var] = {'sources': {f'{op}_{var}_col_{idx}'},
                              'reduce_dim': True}

        # update edge target information
        edge_dict = circuit.edges[edge]
        edge_dict["target_var"] = f'{op}_{var}_col_{idx}/{var}_col_{idx}'

    def go_through_nodes_and_create_mapping_for_their_inputs(self, circuit):

        # (Previously was) Third Stage: Finalize edge vectorization
        ##########################################

        # go through nodes and create mapping for their inputs
        for node_name, data in circuit.nodes(data=True):
            node = data["node"]

            incoming_edges = circuit.graph.in_edges(node_name)
            incoming_edges = self._sort_edges(circuit, incoming_edges, 'target_var')

            # loop over input variables of node
            for in_var, edges in incoming_edges.items():

                # extract info for input variable connections
                n_inputs = len(edges)
                op_name, var_name = in_var.split('/')
                delays = []
                for s, t, e in edges:
                    d = circuit.edges[(s, t, e)]['delay']
                    if d is not None:
                        delays.append(d)

                if delays:
                    max_delay = np.max(delays)
                else:
                    max_delay = None

                # loop over different input sources
                for j in range(n_inputs):

                    if max_delay is not None:

                        # add synaptic buffer to the input variable
                        self._add_edge_buffer(circuit, node, op_name, var_name, idx=j, buffer_length=max_delay,
                                              edge=edges[j])

                    elif n_inputs > 1:

                        # add synaptic input collector to the input variable
                        self._add_edge_input_collector(circuit, node, op_name, var_name, idx=j, edge=edges[j])

    @staticmethod
    def _map_multiple_inputs(inputs: dict, reduce_dim: bool) -> tuple:
        """Creates mapping between multiple input variables and a single output variable.

        Parameters
        ----------
        inputs
            Input variables.
        reduce_dim
            If true, input variables will be summed up, if false, they will be concatenated.
        kwargs
            Additional keyword arguments for the backend.

        Returns
        -------
        tuple
            Summed up or concatenated input variables and the mapping to the respective input variables

        """

        inputs_unique = []
        input_mapping = {}
        for key, var in inputs.items():
            node, in_op, in_var = key.split('/')
            i = 0
            inp = in_var
            while inp in inputs_unique:
                i += 1
                if inp[-2:] == f"_{i - 1}":
                    inp[:-2] = f"_{i}"
                else:
                    inp = f"{inp}_{i}"
            inputs_unique.append(inp)
            input_mapping[inp] = key

        if reduce_dim:
            inputs_unique = f"sum(({','.join(inputs_unique)}), 0)"
        else:
            idx = 0
            var = inputs[input_mapping[inputs_unique[idx]]]
            while not hasattr(var, 'shape'):
                idx += 1
                var = inputs[input_mapping[inputs_unique[idx]]]
            shape = var['shape']
            if len(shape) > 0:
                inputs_unique = f"reshape(({','.join(inputs_unique)}), ({len(inputs_unique) * shape[0],}))"
            else:
                inputs_unique = f"stack({','.join(inputs_unique)})"
        return inputs_unique, input_mapping


def sort_equations(edge_eqs: list, node_eqs: list) -> list:
    """

    Parameters
    ----------
    edge_eqs
    node_eqs

    Returns
    -------

    """

    from pyrates.backend.parser import is_diff_eq

    # clean up equations
    for i, layer in enumerate(edge_eqs.copy()):
        if not layer:
            edge_eqs.pop(i)
    for i, layer in enumerate(node_eqs.copy()):
        if not layer:
            node_eqs.pop(i)

    # re-order node equations
    eqs_new = []
    for node_layer in node_eqs.copy():
        if not any([is_diff_eq(eq) for eq, _ in node_layer]):
            eqs_new.append(node_layer)
            node_eqs.pop(node_eqs.index(node_layer))

    eqs_new += edge_eqs
    eqs_new += node_eqs

    return eqs_new



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network models and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation
"""Functionality to register functions that are used to transform between different data types in the frontend.
"""

__author__ = "Daniel Rose"
__status__ = "Development"

REGISTERED_INTERFACES = dict()


def register_interface(func, name=""):
    """Register a transformation function (interface) between two representations of models.

    Parameters
    ----------
    func
        Function to be registered. Needs to start with "from_" or "to_" to signify the direction of transformation
    name
        (Optional) String that defines the name under which the function should be registered. If left empty,
        the name will be formatted in the form {target}_from_{source}, where target and source are representations to
        transform from or to."""
    if name is "":

        # get interface name from module name
        module_name = func.__module__.split(".")[-1]
        # parse to_ and from_ functions
        func_name = func.__name__
        if func_name.startswith("from_"):
            target = module_name
            source = func_name[5:]  # crop 'from_'
        elif func_name.startswith("to_"):
            source = module_name
            target = func_name[3:]  # crop 'to_'
        else:
            raise ValueError(f"Function name {func_name} does not adhere to convention to start "
                             f"with either `to_` or `from_`.")  # ignore any other functions
        new_name = f"{target}_from_{source}"
    else:
        new_name = name

    if new_name in REGISTERED_INTERFACES:
        raise ValueError(f"Interface {new_name} already exist. Cannot add {func}.")
    else:
        REGISTERED_INTERFACES[new_name] = func

    return func



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation


#############################################################################################################
# Define shortcuts for different types of input that are defined as frontends to provide a common interface
# to all implemented type conversions.
# All functions are renamed explicitly
#############################################################################################################

# template-based interface
from pyrates.frontend import template
# from pyrates.frontend import dict as dict_
from pyrates.frontend import yaml
# from pyrates.frontend import nxgraph
from pyrates.frontend.template import CircuitTemplate, NodeTemplate, EdgeTemplate, OperatorTemplate

# By importing the above, all transformation functions (starting with `to_` or `from_`) are registered
# Below these functions are collected and made available from pyrates.frontend following the naming convention
# `{target}_from_{source}` with target and source being respective valid representations in the frontend

from pyrates.frontend._registry import REGISTERED_INTERFACES
import sys

# add all registered functions to main frontend module
this_module = sys.modules[__name__]

for new_name, func in REGISTERED_INTERFACES.items():
    # set new name on current module
    setattr(this_module, new_name, func)


# The following function are shorthands that bridge multiple interface steps

def circuit_from_yaml(path: str):
    """Directly return CircuitIR instance from a yaml file."""
    return CircuitTemplate.from_yaml(path).apply()


def node_from_yaml(path: str):
    """Directly return NodeIR instance from a yaml file."""
    return NodeTemplate.from_yaml(path).apply()


def edge_from_yaml(path: str):
    """Directly return EdgeIR instance from a yaml file."""

    return EdgeTemplate.from_yaml(path).apply()


def operator_from_yaml(path: str):
    """Directly return OperatorIR instance from a yaml file."""

    return OperatorTemplate.from_yaml(path).apply()




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from copy import deepcopy

import networkx as nx
from networkx import MultiDiGraph, DiGraph, find_cycle, NetworkXNoCycle

from pyrates import PyRatesException
from pyrates.ir.edge import EdgeIR
from pyrates.ir.circuit import CircuitIR
from pyrates.frontend.dict import to_node
from pyrates.frontend._registry import register_interface


__author__ = "Daniel Rose"
__status__ = "Development"


# @register_interface
# def to_circuit(graph: nx.MultiDiGraph, label="circuit",
#                node_creator=to_node):
#     """Create a CircuitIR instance out of a networkx.MultiDiGraph"""
#
#     circuit = CircuitIR(label)
#
#     for name, data in graph.nodes(data=True):
#
#         circuit.add_node(name, node=node_creator(data))
#
#     required_keys = ["source_var", "target_var", "weight", "delay"]
#     for source, target, data in graph.edges(data=True):
#
#         if all([key in data for key in required_keys]):
#             if "edge_ir" not in data:
#                 data["edge_ir"] = EdgeIR()
#             source_var = data.pop("source_var")
#             target_var = data.pop("target_var")
#             circuit.add_edge(f"{source}/{source_var}", f"{target}/{target_var}", **data)
#         else:
#             raise KeyError(f"Missing a key out of {required_keys} in an edge with source `{source}` and target"
#                            f"`{target}`")
#
#     return circuit
#
#
# def from_circuit(circuit, revert_node_names=False):
#     """Old implementation that transforms all information in a circuit to a networkx.MultiDiGraph with a few additional
#     transformations that the old backend needed."""
#     return Circuit2NetDef.network_def(circuit, revert_node_names)
#
#
# class Circuit2NetDef:
#     # label_counter = {}  # type: Dict[str, int]
#
#     @classmethod
#     def network_def(cls, circuit: CircuitIR, revert_node_names=False):
#         """A bit of a workaround to connect interfaces of frontend and backend.
#         TODO: Remove BackendIRFormatter and adapt corresponding tests"""
#         # import re
#
#         network_def = MultiDiGraph()
#
#         edge_list = []
#         node_dict = {}
#
#         # reorganize node to conform with backend API
#         #############################################
#         for node_key, data in circuit.graph.nodes(data=True):
#             node = data["node"]
#             # reformat all node internals into operators + operator_args
#             if revert_node_names:
#                 names = node_key.split("/")
#                 node_key = ".".join(reversed(names))
#             node_dict[node_key] = {}  # type: Dict[str, Union[list, dict]]
#             node_dict[node_key] = dict(cls._nd_reformat_operators(node.op_graph))
#             op_order = cls._nd_get_operator_order(node.op_graph)  # type: list
#             # noinspection PyTypeChecker
#             node_dict[node_key]["operator_order"] = op_order
#
#         # reorganize edge to conform with backend API
#         #############################################
#         for source, target, data in circuit.graph.edges(data=True):
#             # move edge operators to node
#             if revert_node_names:
#                 source = ".".join(reversed(source.split("/")))
#                 target = ".".join(reversed(target.split("/")))
#             node_dict[target], edge = cls._move_edge_ops_to_node(target, node_dict[target], data)
#
#             edge_list.append((source, target, dict(**edge)))
#
#         # network_def.add_nodes_from(node_dict)
#         for key, node in node_dict.items():
#             network_def.add_node(key, **node)
#         network_def.add_edges_from(edge_list)
#
#         return network_def  # return MultiDiGraph as needed by ComputeGraph class
#
#     @staticmethod
#     def _nd_reformat_operators(op_graph: DiGraph):
#         operator_args = dict()
#         operators = dict()
#
#         for op_key, op_dict in op_graph.nodes(data=True):
#             op_cp = deepcopy(op_dict)  # duplicate operator info
#             var_dict = op_cp["operator"].variables
#             for var_key, var_props in var_dict.items():
#                 operator_args[f"{op_key}/{var_key}"] = var_props
#
#             op_cp["equations"] = op_cp["operator"].equations
#             op_cp["inputs"] = op_cp["operator"].inputs
#             op_cp["output"] = op_cp["operator"].output
#             # op_cp.pop("values", None)
#             op_cp.pop("operator", None)
#             operators[op_key] = op_cp
#
#         reformatted = dict(operator_args=operator_args,
#                            operators=operators,
#                            inputs={})
#         return reformatted
#
#     @staticmethod
#     def _nd_get_operator_order(op_graph: DiGraph) -> list:
#         """
#
#         Parameters
#         ----------
#         op_graph
#
#         Returns
#         -------
#         op_order
#         """
#         # check, if cycles are present in operator graph (which would be problematic
#         try:
#             find_cycle(op_graph)
#         except NetworkXNoCycle:
#             pass
#         else:
#             raise PyRatesException("Found cyclic operator graph. Cycles are not allowed for operators within one node.")
#
#         op_order = []
#         graph = op_graph.copy()  # type: DiGraph
#         while graph.nodes:
#             # noinspection PyTypeChecker
#             primary_nodes = [node for node, in_degree in graph.in_degree if in_degree == 0]
#             op_order.extend(primary_nodes)
#             graph.remove_nodes_from(primary_nodes)
#
#         return op_order
#
#     @classmethod
#     def _move_edge_ops_to_node(cls, target, node_dict: dict, edge_dict: dict) -> (dict, dict):
#         """
#
#         Parameters
#         ----------
#         target
#             Key identifying target node in backend graph
#         node_dict
#             Dictionary of target node (to move operators into)
#         edge_dict
#             Dictionary with edge properties (to move operators from)
#         Returns
#         -------
#         node_dict
#             Updated dictionary of target node
#         edge_dict
#              Dictionary of reformatted edge
#         """
#         # grab all edge variables
#         edge = edge_dict["edge_ir"]  # type: EdgeIR
#         source_var = edge_dict["source_var"]
#         target_var = edge_dict["target_var"]
#         weight = edge_dict["weight"]
#         delay = edge_dict["delay"]
#         input_var = edge.input
#         output_var = edge.output
#
#         if len(edge.op_graph) > 0:
#             # reformat all edge internals into operators + operator_args
#             op_data = cls._nd_reformat_operators(edge.op_graph)  # type: dict
#             op_order = cls._nd_get_operator_order(edge.op_graph)  # type: List[str]
#             operators = op_data["operators"]
#             operator_args = op_data["operator_args"]
#
#             # operator keys refer to a unique combination of template names and changed values
#
#             # add operators to target node in reverse order, so they can be safely prepended
#             added_ops = False
#             for op_name in reversed(op_order):
#                 # check if operator name is already known in target node
#                 if op_name in node_dict["operators"]:
#                     pass
#                 else:
#                     added_ops = True
#                     # this should all go smoothly, because operator should not be known yet
#                     # add operator dict to target node operators
#                     node_dict["operators"][op_name] = operators[op_name]
#                     # prepend operator to op_order
#                     node_dict["operator_order"].insert(0, op_name)
#                     # ToDo: consider using collections.deque instead
#                     # add operator args to target node
#                     node_dict["operator_args"].update(operator_args)
#
#             out_op = op_order[-1]
#             out_var = operators[out_op]['output']
#             if added_ops:
#                 # append operator output to target operator sources
#                 # assume that only last operator in edge operator_order gives the output
#                 # for op_name in node_dict["operators"]:
#                 #     if out_var in node_dict["operators"][op_name]["inputs"]:
#                 #         if out_var_long not in node_dict["operators"][op_name]["inputs"][out_var]:
#                 #             # add reference to source operator that was previously in an edge
#                 #             node_dict["operators"][op_name]["inputs"][out_var].append(output_var)
#
#                 # shortcut, since target_var and output_var are known:
#                 target_op, target_vname = target_var.split("/")
#                 if output_var not in node_dict["operators"][target_op]["inputs"][target_vname]["sources"]:
#                     node_dict["operators"][target_op]["inputs"][target_vname]["sources"].append(out_op)
#
#             # simplify edges and save into edge_list
#             # op_graph = edge.op_graph
#             # in_ops = [op for op, in_degree in op_graph.in_degree if in_degree == 0]
#             # if len(in_ops) == 1:
#             #     # simple case: only one input operator? then it's the first in the operator order.
#             #     target_op = op_order[0]
#             #     target_inputs = operators[target_op]["inputs"]
#             #     if len(target_var) != 1:
#             #         raise PyRatesException("Either too many or too few input variables detected. "
#             #                                "Needs to be exactly one.")
#             #     target_var = list(target_inputs.keys())[0]
#             #     target_var = f"{target_op}/{target_var}"
#             # else:
#             #     raise NotImplementedError("Transforming an edge with multiple input operators is not yet handled.")
#
#             # shortcut to new target war:
#             target_var = input_var
#         edge_dict = {"source_var": source_var,
#                      "target_var": target_var,
#                      "weight": weight,
#                      "delay": delay}
#         # set target_var to singular input of last operator added
#         return node_dict, edge_dict



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Utility functions to store Circuit configurations and data in JSON files
and read/construct circuit from JSON.
"""

# external packages
from collections import OrderedDict
from typing import Generator, Tuple, Any, Union, List, Dict
from networkx import node_link_data
from inspect import getsource
import numpy as np
import json
from pandas import DataFrame
import pandas as pd

# meta infos
__author__ = "Daniel Rose"
__status__ = "Development"

# TODO: Update documentations & clean functions from unnecessary comments (i.e. silent code)


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):

        # from pyrates.population import Population
        # from pyrates.synapse import Synapse

        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif hasattr(obj, "to_json"):
            return obj.to_dict()
        # elif isinstance(obj, Synapse):
        #     attr_dict = obj.to_dict()
        #     # remove synaptic_kernel from dict, if kernel_function is specified
        #     if "kernel_function" in attr_dict:
        #       JansenRitCircuit  attr_dict.pop("synaptic_kernel", None)
        #     return attr_dict
        # elif isinstance(obj, Axon):
        #     return get_attrs(obj)
        elif callable(obj):
            return getsource(obj)
        else:
            return super().default(obj)


class RepresentationBase(object):
    """Class that implements a __repr__ that yields the __init__ function signature with provided arguments in the
    form 'module.Class(arg1, arg2, arg3)'"""

    def __new__(cls, *args, **kwargs):

        # noinspection PyArgumentList
        instance = super().__new__(cls)
        _init_dict = {**kwargs}
        if args:
            _init_dict["args"] = args

        instance._init_dict = _init_dict

        return instance

    def __repr__(self) -> str:
        """Magic method that returns to repr(object). It retrieves the signature of __init__ and maps it to class
        attributes to return a representation in the form 'module.Class(arg1=x, arg2=y, arg3=3)'. Raises AttributeError,
        if the a parameter in the signature cannot be found (is not saved). The current implementation """

        # repr_str = f"{self.__module__!r}.{self.__class__!r}("
        # params = self._get_params()
        from copy import copy
        init_dict = copy(self._init_dict)

        # args = ""
        # if "args" in init_dict:
        #     args = init_dict.pop("args", None)
        #     args = ", ".join((f"{value!r}" for value in args))
        #
        # kwargs = ", ".join((f"{name}={value!r}" for name, value in init_dict.items()))
        # param_str = f"{args}, {kwargs}"
        param_str = "*args, **kwargs"
        _module = self.__class__.__module__
        _class = self.__class__.__name__
        # return f"{_module}.{_class}({param_str})"

        try:
            return f"{_class} '{self.key}'"
        except AttributeError:
            return f"{_class}"

    def _defaults(self) -> Generator[Tuple[str, Any], None, None]:
        """Inspects the __init__ special method and yields tuples of init parameters and their default values."""

        import inspect

        # retrieve signature of __init__ method
        sig = inspect.signature(self.__init__)

        # retrieve parameters in the signature
        for name, param in sig.parameters.items():
            # check if param has a default value
            if np.all(param.default != inspect.Parameter.empty):
                # yield parameter name and default value
                yield name, param.default

    def to_dict(self, include_defaults=False, include_graph=False, recursive=False) -> OrderedDict:
        """Parse representation string to a dictionary to later convert it to json."""

        _dict = OrderedDict()
        _dict["class"] = {"__module__": self.__class__.__module__, "__name__": self.__class__.__name__}
        # _dict["module"] = self.__class__.__module__
        # _dict["class"] = self.__class__.__name__

        # obtain parameters that were originally passed to __init__
        ###########################################################

        for name, value in self._init_dict.items():
            _dict[name] = value

        # include defaults of parameters that were not specified when __init__ was called.
        ##################################################################################

        if include_defaults:
            default_dict = {}
            for name, value in self._defaults():
                if name not in _dict:  # this fails to get positional arguments that override defaults to keyword args
                    default_dict[name] = value
            _dict["defaults"] = default_dict

        # Include graph if desired
        ##########################

        # check if it quacks like a duck... eh.. has a network_graph like a circuit
        if include_graph and hasattr(self, "network_graph"):
            net_dict = node_link_data(self.network_graph)
            if recursive:
                for node in net_dict["nodes"]:
                    node["data"] = node["data"].to_dict(include_defaults=include_defaults, recursive=True)
            _dict["network_graph"] = net_dict

        # Apply dict transformation recursively to all relevant objects
        ###############################################################

        if recursive:
            for key, item in _dict.items():
                # check if it quacks like a duck... eh.. represents like a RepresentationBase
                # note, however, that "to_dict" is actually implemented by a number of objects outside this codebase
                if hasattr(item, "to_dict"):
                    _dict[key] = item.to_dict(include_defaults=include_defaults, recursive=True)

        return _dict

    def to_json(self, include_defaults=False, include_graph=False, path="", filename=""):
        """Parse a dictionary into """

        # from pyrates.utility.json_filestorage import CustomEncoder

        _dict = self.to_dict(include_defaults=include_defaults, include_graph=include_graph, recursive=True)

        if filename:
            import os
            filepath = os.path.join(path, filename)

            # create directory if necessary
            create_directory(filepath)

            if not filepath.lower().endswith(".json"):
                filepath = f"{filepath}.json"

            with open(filepath, "w") as outfile:
                json.dump(_dict, outfile, cls=CustomEncoder, indent=4)

        # return json as string
        return json.dumps(_dict, cls=CustomEncoder, indent=2)


def get_simulation_data(circuit, state_variable='membrane_potential', pop_keys: Union[tuple, list] = None,
                        time_window: tuple = None) -> Tuple[dict, DataFrame]:
    """Obtain all simulation data from a circuit, including run parameters"""

    run_info = circuit.run_info
    states = circuit.get_population_states(state_variable=state_variable, population_keys=pop_keys,
                                           time_window=time_window)

    labels = [pop for pop in circuit.populations.keys() if 'dummy' not in pop]

    states = DataFrame(data=states, index=run_info.index, columns=labels)

    return run_info, states


def save_simulation_data_to_file(output_data: DataFrame, run_info: dict,
                                 dirname: str, path: str = "", out_format: str = "csv"):
    """Save simulation output and inputs that were given to the run function to a file."""

    import os

    dirname = os.path.join(path, dirname) + "/"

    # create directory if necessary
    create_directory(dirname)

    # save output data
    ##################

    filename = f"output.{out_format}"
    filepath = os.path.join(dirname, filename)
    if out_format == "json":
        output_data.to_json(filepath, orient="split")
    elif out_format == "csv":
        output_data.to_csv(filepath, sep="\t")
    else:
        raise ValueError(f"Unknown output format '{out_format}'")

    # save run information
    ######################

    if "time_vector" in run_info:
        # throw away time vector, since it is contained in all the other data
        run_info.pop("time_vector")

    for key, item in run_info.items():
        filename = f"{key}.{out_format}"
        filepath = os.path.join(dirname, filename)
        if item is None:
            continue

        if out_format == "json":
            item.to_json(filepath, orient="split")
        else:
            item.to_csv(filepath, sep="\t")

    # TODO: think about float vs. decimal representation: Possibly save float data as hex or fraction
    # TODO: choose time update as exact float (e.g. 0.0005 --> 2**-11=0.00048828125)


def create_directory(path):
    """check if a directory exists and create it otherwise"""

    import os
    import errno

    if not os.path.exists(os.path.dirname(path)):
        try:
            os.makedirs(os.path.dirname(path))
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise


def read_simulation_data_from_file(dirname: str, path="", filenames: list = None) -> Dict[str, DataFrame]:
    """Read simulation data from files. The assumed data structure is:
    <path>/<dirname>/
        output.csv
        synaptic_inputs.csv
        extrinsic_current.csv
        extrinsic_modulation.csv

    This is the expected output from 'save_simulation_data_to_file', but if 'names' is specified, other data
    may also be read.
    """

    import os

    if filenames is None:
        filenames = ["output", "synaptic_inputs", "extrinsic_current", "extrinsic_modulation"]
        ignore_missing = True
    else:
        ignore_missing = False

    data = {}
    path = os.path.join(path, dirname)

    for label in filenames:
        if label == "output":
            header = 0
        else:
            header = [0, 1]
        filename = label + ".csv"
        filepath = os.path.join(path, filename)

        try:
            data[label] = pd.read_csv(filepath, sep="\t", header=header, index_col=0)
        except FileNotFoundError:
            if not ignore_missing:
                raise

    return data




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""Provides functions to build MNE objects (raw, epoch, evoked) from PyRates simulation results (csv) or from circuit
objects.
"""

# external packages
import numpy as np
from typing import Union, Optional, List, Any
from pandas import DataFrame

# pyrates internal imports

# meta infos
__author__ = 'Richard Gast'
__status__ = 'Development'


############################################################
# wrapper to build an mne object from observer information #
############################################################


def mne_from_dataframe(sim_results: DataFrame,
                       ch_types: Union[str, List[str]] = 'eeg',
                       ch_names: Optional[Union[str, List[str]]] = None,
                       events: Optional[np.ndarray] = None,
                       event_keys: Optional[List[str]] = None,
                       epoch_start: Optional[float] = None,
                       epoch_end: Optional[float] = None,
                       epoch_duration: Optional[float] = None,
                       epoch_averaging: bool = False,
                       ) -> Any:
    """Uses the data stored on a circuit to create a raw/epoch/evoked mne object.

    Parameters
    ----------
    sim_results
        Pandas dataframe in which circuit simulation results are stored (output of circuit's `run` function).
    ch_types
        Type of the channels, the observation time-series of the observers refer to.
    ch_names
        Name of each channel/observation time-series.
    events
        2D array defining events during the simulation. For a more detailed documentation, see the docstring for
        parameter `events` of :class:`mne.Epochs`.
    event_keys
        Names of the events. For a more detailed documentation, see the docstring for parameter `event_id` of
        :class:`mne.Epochs`.
    epoch_start
        Time, relative to event onset, at which an epoch should start [unit = s]. For a more detailed documentation,
        see the docstring for parameter `tmin` of :class:`mne.Epochs`.
    epoch_end
        Time, relative to event onset, at which an epoch should end [unit = s]. For a more detailed documentation,
        see the docstring for parameter `tmax` of :class:`mne.Epochs`.
    epoch_duration
        Instead of passing `events`, this parameter can be used to create epochs with a fixed duration [unit = s].
        If this is used, do not pass `epoch_start` or `epoch_end`. For a more detailed documentation,
        see the docstring for parameter `duration` of :function:`mne.make_fixed_length_events`.
    epoch_averaging
        Only relevant, if `events` or `event_duration` were passede. If true, an :class:`mne.EvokedArray` instance will
        be returned that contains ttime-series averaged over all epochs.

    Returns
    -------
    Any
        MNE object that contains either the raw, epoched, or averaged (over epochs) data.

    """

    import mne

    # extract information from arguments
    ####################################

    dt = sim_results.index[1] - sim_results.index[0]

    # channel information
    if not ch_names:
        ch_names = list(sim_results.keys())
    if type(ch_types) is str:
        ch_types = [ch_types for _ in range(len(ch_names))]
    ch_names_str = []
    for name in ch_names:
        ch_names_str.append(str(name))

    # epoch/event information
    if not epoch_start:
        epoch_start = -0.2 if not epoch_duration else 0.
    if not epoch_end:
        epoch_end = 0.5 if not epoch_duration else epoch_duration - 1/dt

    # create raw mne object
    #######################

    # create mne info object
    info = mne.create_info(ch_names=ch_names_str, ch_types=ch_types, sfreq=1/dt)

    # create raw object from info and circuit data
    raw = mne.io.RawArray(data=sim_results[ch_names].values.T, info=info)

    # create final mne object
    #########################

    # if events or event information is passed, create Epoch or Evoked object
    if events is not None or epoch_duration:

        # check whether events still have to be created or not
        if events is None:
            events = mne.make_fixed_length_events(raw=raw, id=0, duration=epoch_duration)

        # check whether event labels still have to be created or not
        if not event_keys:
            event_keys = dict()
            for event in np.unique(events[:, 2]):
                event_keys['event_' + str(event)] = event

        # create epoch object from raw data and event information
        mne_object = mne.Epochs(raw=raw, events=events, event_id=event_keys, tmin=epoch_start, tmax=epoch_end)

        # create Evoked object by averaging over epochs if epoch_averaging is true
        if epoch_averaging:

            # average over epoch data
            data = mne_object.get_data()
            n_epochs = len(data)
            data = np.mean(data, axis=0)

            # create evoked object
            mne_object = mne.EvokedArray(data=data, info=info, tmin=epoch_start, comment=event_keys['event_0'],
                                         nave=n_epochs)

    # stick with Raw object
    else:

        mne_object = raw

    return mne_object


###################################################
# wrapper to build an mne object from output file #
###################################################


def mne_from_csv(csv_dir: str,
                 ch_types: Union[str, List[str]] = 'eeg',
                 ch_names: Optional[Union[str, List[str]]] = None,
                 events: Optional[np.ndarray] = None,
                 event_keys: Optional[List[str]] = None,
                 epoch_start: Optional[float] = None,
                 epoch_end: Optional[float] = None,
                 epoch_duration: Optional[float] = None,
                 epoch_averaging: bool = False,
                 ) -> Any:
    """Uses the data stored on circuit to create a raw/epoch/evoked mne object.

    Parameters
    ----------
    csv_dir
        Full path + filename of the csv file that contains the circuit outputs from which an MNE object should be
        created.
    ch_types
        Type of the channels, the observation time-series of the observers refer to.
    ch_names
        Name of each channel/observation time-series.
    events
        2D array defining events during the simulation. For a more detailed documentation, see the docstring for
        parameter `events` of :class:`mne.Epochs`.
    event_keys
        Names of the events. For a more detailed documentation, see the docstring for parameter `event_id` of
        :class:`mne.Epochs`.
    epoch_start
        Time, relative to event onset, at which an epoch should start [unit = s]. For a more detailed documentation,
        see the docstring for parameter `tmin` of :class:`mne.Epochs`.
    epoch_end
        Time, relative to event onset, at which an epoch should end [unit = s]. For a more detailed documentation,
        see the docstring for parameter `tmax` of :class:`mne.Epochs`.
    epoch_duration
        Instead of passing `events`, this parameter can be used to create epochs with a fixed duration [unit = s].
        If this is used, do not pass `epoch_start` or `epoch_end`. For a more detailed documentation,
        see the docstring for parameter `duration` of :function:`mne.make_fixed_length_events`.
    epoch_averaging
        Only relevant, if `events` or `event_duration` were passede. If true, an :class:`mne.EvokedArray` instance will
        be returned that contains ttime-series averaged over all epochs.

    Returns
    -------
    Any
        MNE object that contains either the raw, epoched, or averaged (over epochs) data.

    """

    import mne

    # extract information from arguments
    ####################################

    # from data file
    from pandas.io.parsers import read_csv

    output = read_csv(csv_dir, delim_whitespace=True, header=0)
    states = output.values
    sfreq = states[-1, 0] / states.shape[0]
    states = states[:, 1:]
    n_channels = states.shape[1]

    # channel information
    if type(ch_types) is str:
        ch_types = [ch_types for _ in range(n_channels)]
    if not ch_names:
        ch_names = output.keys()
        ch_names.pop(0)

    # epoch/event information
    if not epoch_start:
        epoch_start = -0.2 if not epoch_duration else 0.
    if not epoch_end:
        epoch_end = 0.5 if not epoch_duration else epoch_duration - 1 / sfreq

    # create raw mne object
    #######################

    # create mne info object
    info = mne.create_info(ch_names=ch_names, ch_types=ch_types, sfreq=sfreq)

    # create raw object from info and circuit data
    raw = mne.io.RawArray(data=states.T, info=info)

    # create final mne object
    #########################

    # if events or event information is passed, create Epoch or Evoked object
    if events is not None or epoch_duration:

        # check whether events still have to be created or not
        if events is None:
            events = mne.make_fixed_length_events(raw=raw, id=0, duration=epoch_duration)

        # check whether event labels still have to be created or not
        if not event_keys:
            event_keys = dict()
            for event in np.unique(events[:, 2]):
                event_keys['event_' + str(event)] = event

        # create epoch object from raw data and event information
        mne_object = mne.Epochs(raw=raw, events=events, event_id=event_keys, tmin=epoch_start, tmax=epoch_end)

        # create Evoked object by averaging over epochs if epoch_averaging is true
        if epoch_averaging:

            # average over epoch data
            data = mne_object.get_data()
            n_epochs = len(data)
            data = np.mean(data, axis=0)

            # create evoked object
            mne_object = mne.EvokedArray(data=data, info=info, tmin=epoch_start, comment=event_keys['event_0'],
                                         nave=n_epochs)

    # stick with Raw object
    else:

        mne_object = raw

    return mne_object
